#!/usr/bin/python
# coding=UTF-8
# 
# bulk_extractor.py:
# A module for working with bulk_extractor


__version__ = "1.0.0"

import zipfile,os,os.path,glob

class BulkReport:
    """Creates an object from a bulk_extractor report. The report can be a directory or a ZIP of a directory.
    Methods that you may find useful:
    f = b.open(fname,mode) - opens the file f and returns a file handle. mode defaults to 'rb'
    b.is_histogram_file(fname) - returns if fname is a histogram file or not
    b.image_file() - Returns the name of the image file
    b.read_histogram(fn) - Reads a histogram and returns the histogram
    b.histograms      - Set of histogram names
    b.files   - Set of feature file names
"""    

    def __init__(self,fn):
        def validate():
            """Validates the XML and finds the histograms and feature files"""
            import xml.dom.minidom
            import xml.parsers.expat
            try:
                self.xmldoc = xml.dom.minidom.parse(self.open("report.xml"))
            except xml.parsers.expat.ExpatError as e:
                print("%s does not contain a valid report.xml file" % (self.name))
                print(e)
                exit(1)
            return True

        import os.path,glob
        self.name  = fn
        if os.path.isdir(fn):
            self.zipfile = None
            self.dname = fn
            self.all_files = set([os.path.basename(x) for x in glob.glob(os.path.join(fn,"*"))])
            self.files = set([os.path.basename(x) for x in glob.glob(os.path.join(fn,"*.txt"))])
            validate()
            return
        if fn.endswith(".zip") and os.path.isfile(fn):
            self.zipfile = zipfile.ZipFile(fn)
            # extract the filenames and make a map
            self.files = set()
            self.map   = dict()
            for fn in self.zipfile.namelist():
                self.files.add(os.path.basename(fn))
                self.map[os.path.basename(fn)] = fn
            validate()
            return
        raise RuntimeError("Cannot process " + fn)

    def imagefile(self):
        """Returns the file name of the disk image that was used."""
        return self.xmldoc.getElementsByTagName("image_filename")[0].firstChild.wholeText

    def version(self):
        """Returns the version of bulk_extractor that made the file."""
        return self.xmldoc.getElementsByTagName("version")[0].firstChild.wholeText

    def open(self,fname,mode='rb'):
        """Opens a named file in the bulk report."""
        if self.zipfile:
            return self.zipfile.open(self.map[fname],mode=mode)
        else:
            fn = os.path.join(self.dname,fname)
            if not os.path.exists(fn):
                print("FILE NOT FOUND: %s" % fn)
                exit(1)
            return open(fn,mode=mode)

    def is_histogram_file(self,fname):
        """Returns true if this is a histogram file"""
        if "_histogram" in fname: return True
        if "url_" in fname: return True # I know this is a histogram file
        if fname=="ccn_track2.txt": return False # I know this is not
        # Check the first line
        for line in self.open(fname):
            if line[0:1]=="#":  # can't decide if a comment is a histogram or not
                continue
            if line[0:2]=="n=": # histogram
                return True
            return False
        return False            # no data


    def is_feature_file(self,fname):
        """Returns true if this is a feature file"""
        if "_histogram" in fname: return False
        if "_stopped" in fname: return False
        if self.is_histogram_file(fname): return False
        return True

    def histograms(self):
        """Returns a list of the histograms, by name"""
        return filter(lambda fn:self.is_histogram_file(fn),self.files)

    def feature_files(self):
        """Returns a list of the feature_files, by name"""
        return filter(lambda fn:self.is_feature_file(fn),self.files)

    def read_histogram(self,fn):
        """Read a histogram file and return a dictonary of the histogram """
        import re
        ret = {}
        r = re.compile(b"^n=(\d+)\t(.*)$")
        for line in self.open(fn):
            m = r.search(line)
            if m:
                ret[m.group(2)] = int(m.group(1))
        return ret
        
if(__name__=='__main__'):
    from optparse import OptionParser
    global options
    import os

    parser = OptionParser()
    parser.add_option("--wordlist",help="split a wordlist.txt file")
    (options,args) = parser.parse_args()

    if options.wordlist:
        global counter
        fn = options.wordlist
        if not os.path.exists(fn):
            print("%s does not exist" % fn)
            exit(1)
        print("Splitting wordlist file %s" % fn)
        fntemp = fn.replace(".txt","_%03d.txt")
        counter = 0
        seen = set()

        def next_file():
            global counter
            nfn = fntemp % counter
            counter += 1
            print("Switching to file %s" % nfn) 
            return open(nfn,"w")

        f = next_file()
        size = 0
        max_size = 1000*1000*10
        for line in open(fn,"r"):
            word = line.split("\t")[1]
            if word not in seen:
                f.write(word)
                if word[-1]!='\n': f.write("\n")
                seen.add(word)
                size += len(word)+1
                if size>max_size:
                    f.close()
                    f = next_file()
                    size = 0
