#!/usr/bin/python
#
# CDA tool: reads multiple bulk_extractor histogram files and outputs:
# stoplist.txt - list of email addresses on more than 1/3 of the disks
# targets.txt  - list of email addresses not on stoplist and the # of drives on which they appear.

import os,sys,re
if os.getenv("DOMEX_HOME"): sys.path.append(os.getenv("DOMEX_HOME") + "/src/lib/") # add the library
sys.path.append("../lib/")      # add the library

# This regular expression matches on bulk_extractor offset and item
feature_re  = re.compile("([^\t]+)\t([^\t\n\r]+)") 

def mycmp2(a,b):
    if a[1]>b[1]: return -1
    if a[1]<b[1]: return 1
    return 0


class correlator:
    """The correlator has two dictionary. items:
    key = value being correlated.
    value = dictionary of driveID and the count on each drive."""
    def __init__(self,name):
        self.name = name
        self.drive_to_id = {}        # maps drive names to ids
        self.id_to_drive = {}        # maps drive names to ids
        self.items  = {}        # maps each item to a set of dictionaries, where each has {id:count}
        self.drive_items = {}   # a histogram for each drive of the items that it contains

    def longest_drive_name(self):
        return max([len(s) for s in self.drive_to_id.keys()])

    def longest_item(self):
        return max([len(s) for s in self.items.keys()])

    def get_drive_id(self,drive_name):
        if drive_name not in self.drive_to_id:
            id = len(self.drive_to_id)+1
            self.drive_to_id[drive_name] = id
            self.id_to_drive[id] = drive_name
            self.drive_items[id] = {}
        return self.drive_to_id[drive_name]

    def add_feature_file(self,drive_name,feature_fn):
        drive_id = None
        for line in open(feature_fn):
            m = feature_re.search(line)
            if m:

                # First add the drive to the database
                if not drive_id:
                    drive_id = self.get_drive_id(drive_name)
                    drive_items = self.drive_items[drive_id]

                # Now process the item, adding it to items and to drive_items
                item = m.group(2)

                # Add to items correlator
                if item not in self.items:
                    self.items[item] = {}
                if drive_id not in self.items[item]:
                    self.items[item][drive_id] = 1
                else:
                    self.items[item][drive_id] += 1
                
                if item not in drive_items:
                    drive_items[item] = 1
                else:
                    drive_items[item] += 1


    def print_cooccurances(self):
        print("Total Drives; %d" % (len(self.drive_to_id)))
        print("Unique %s evaluated: %d" % (self.name,len(self.items)))
        print("Drives with %s co-occurances:" % self.name)
        longest = str(self.longest_drive_name())
        print "longest:",longest
        for d in self.drive_to_id.keys():
            print d,len(d)

        for (item,drives) in self.items.iteritems():
            if len(drives)<2:
                continue
            print("%s %s" % (self.name,item))
            for (id,count) in drives.iteritems():
                drive_name   = self.id_to_drive[id]
                drive_unique = len(self.drive_items[id].keys())
                drive_total  = sum(self.drive_items[id].values())
                fmt = "   %-"+longest+"s : count=%d (out of %d unique / %d total)"
                print(fmt % (drive_name,count,drive_unique,drive_total))
            print("")

class mdc:
    def __init__(self):
        self.total_drives = 0   # total drives processed
        self.email_drives = {}  # key=email / value=Set of drives on which it is present
        self.email_re = re.compile("(.*): (\d)")

    def add_file(self,fn):
        self.total_drives += 1
        for line in open(fn,"r"):
            m = self.email_re.search(line)
            if m:
                email = m.group(1).lower() # just work with lowercase
                email = email.replace("%20","")
                count = m.group(2)
                if email not in self.email_drives: self.email_drives[email] = set()
                self.email_drives[email].add(fn)

    def email_max(self):
        ret = 0
        for (e,dlist) in self.email_drives.iteritems():
            ret = max(ret,len(dlist))
        return ret

    def generate_stop_list(self,output_fn):
        emax = self.email_max()
        print "total drives:",self.total_drives
        print "max drives per email:",emax
        top_third = emax * 1/3

        stoplist = set()
        out = []
        for (e,dlist) in self.email_drives.iteritems():
            if len(dlist) > top_third:
                out.append((e,len(dlist)))
                stoplist.add(e)
        self.stoplist = stoplist

        out.sort(mycmp2)
        f = open(output_fn,"w")
        for (email,count) in out:
            f.write("%s\t%d\n" % (email,count))
        f.close()

    def generate_datasphere(self,fn):
        dsph = []
        for (e,dlist) in self.email_drives.iteritems():
            if e not in self.stoplist:
                count = len(dlist)
                dsph.append((e,count))
        dsph.sort()
        f = open(fn,"w")
        for (e,count) in dsph:
            f.write("%s\t%d\n" % (e,count))
        f.close()
        self.dsph = dsph

    def paircount(self,fn):
        paircount = {}
        for (e,dlist) in self.email_drives.iteritems():
            if e in self.stoplist: continue
            if len(dlist)==2:
                for drive in dlist:
                    paircount[drive] = paircount.get(drive,0)+1
        for (drive,count) in paircount.iteritems():
            print drive,count
        

    def correlate(self,fn):
        cda_tally = {}          # the score for each drive
        drive_emails = {}       # for each drive, the correlated emails
        for line in open(fn,"r"):
            m = self.email_re.search(line)
            if m:
                email = m.group(1).lower() # just work with lowercase
                email = email.replace("%20","")
                count = m.group(2)
                if email in self.stoplist: continue
                if email not in self.email_drives: continue
                drives = self.email_drives[email]
                if len(drives)==1: continue
                for drive in drives:
                    cda_tally[drive] = cda_tally.get(drive,0) + 1.0 / len(drives)
                    if drive not in drive_emails: drive_emails[drive] = set()
                    drive_emails[drive].add(email)
                    
        drives = cda_tally.keys()
        drive_scores = []
        for drive in drives:
            drive_scores.append((drive,cda_tally[drive]))
        drive_scores.sort(mycmp2)
        for (drive,drive_score) in drive_scores:
            print drive,drive_score
            correlated_scores = []
            for correlating_email in drive_emails[drive]:
                correlated_scores.append((correlating_email,1.0 / len(self.email_drives[correlating_email])))
            correlated_scores.sort(mycmp2)
            for (email,score) in correlated_scores:
                print "    ",email,score
            print ""
                

if(__name__=="__main__"):
    import xml.parsers.expat
    from xml.dom.minidom import parse,parseString

    from optparse import OptionParser
    global options
    parser = OptionParser()
    parser.usage = "usage: %prog [otions] <root>"
    parser.add_option("-c","--corr",dest="target", help="Perform correlation")
    parser.add_option("-d","--debug",action="store_true")
    parser.add_option("--ether",help="Just perform ethernet correlation",action="store_true")
    (options,args) = parser.parse_args()

    root = args[0]

    ether_mdc = correlator("Ethernet MAC")
    for (dirpath,dirnames,filenames) in os.walk(root):
        report_xml = os.path.join(dirpath,"report.xml")
        if os.path.exists(report_xml):
            try:
                doc = parseString(open(report_xml).read())
                image_filename = doc.getElementsByTagName("image_filename")[0].firstChild.data
            except xml.parsers.expat.ExpatError:
                print("Invalid XML file: %s" % report_xml)
                continue
                
            print("Processing %s" % image_filename)
            if options.ether:
                ether_mdc.add_feature_file(image_filename,os.path.join(dirpath,"ether.txt"))
    
    if options.ether:
        print("Ethernet:")
        ether_mdc.print_cooccurances()


    exit(0)
    

    for fn in args:
        mymdc.add_file(fn)

    mymdc.generate_stop_list("stoplist.txt")
    mymdc.generate_datasphere("datasphere.txt")
    #mymdc.paircount("paircount.txt")

    if options.target:
        mymdc.correlate(options.target)

