/*
 * bulk_extractor.cpp:
 * Feature Extraction Framework...
 *
 */

#include "bulk_extractor.h"
#include "aftimer.h"
#include "image_process.h"
#include "threadpool.h"
#include "xml.h"

#include <dirent.h>
#include <ctype.h>
#include <fcntl.h>
#include <set>
#include <setjmp.h>
#include <vector>
#include <queue>
#include <unistd.h>
#include <ctype.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SYS_SYSCTL_H
#include <sys/sysctl.h>
#endif

using namespace std;

std::string HTTP_EOL = "\r\n";		// stdout is in binary form

/* hack for subverson */
std::string svn_date("$Date: 2012-03-09 22:14:49 -0500 (Fri, 09 Mar 2012) $");
std::string svn_revision("$Revision: 8561 $");
std::string svn_author("$Author: simsong $");
std::string svn_headurl("$HeadURL: https://domex.nps.edu/domex/svn/src/bulk_extractor/tags/1.2.x/src/bulk_extractor.cpp $");
std::string svn_id("$Id: bulk_extractor.cpp 8561 2012-03-10 03:14:49Z simsong $");

std::string svn_revision_clean()
{
    string svn_r;
    for(string::const_iterator it = svn_revision.begin();it!=svn_revision.end();it++){
	if(*it>='0' && *it<='9') svn_r.push_back(*it);
    }
    return svn_r;
}

/**
 * debug flags:
 * 1  - print pages as they are seen for each scanner.
 * 8  - do not run any of the scanners (but print the sbufs)
 * 16 - dump data as it is seen
 * 32 - fail malloc every 1000 on histogram counter
 * 1000  - just print size of volume and exit.
 * 1001  - MD5 test
 * 1002  - Allocate 512MB of RAM to simulate low memory conditions.
 */
int   debug=0;
int   opt_quiet = 0;
int   max_sbuf_alloc_errors = 60;
int   max_zero_pages = 100;
int64_t opt_page_start = 0;

/**
 * Singletons:
 * histograms_t histograms - the set of histograms to make
 * feature_recorder_set fs - the collection of feature recorders.
 * xml xreport             - the DFXML output.
 * image_process p         - the image being processed.
 * 
 * Note that all of the singletons are passed to the phase1() function.
 */


/** Name of feature files that should be histogramed.
 * The histogram should be done in the plug-in
 */
histograms_t histograms;

#ifdef WIN32 
// Allows us to open standard input in binary mode by default 
// See http://gnuwin32.sourceforge.net/compile.html for more 
int _CRT_fmode = _O_BINARY;
#endif

#ifndef HAVE_LOCALTIME_R
/* locking localtime_r implementation */
void localtime_r(time_t *t,struct tm *tm)
{
}
#endif

#ifndef HAVE_GMTIME_R
/* locking localtime_r implementation */
void gmtime_r(time_t *t,struct tm *tm)
{
}
#endif



/****************************************************************
 *** SCANNER PLUG-IN SYSTEM
 ****************************************************************/

/* scanner_def is the class that is used internally to track all of the plug-in scanners.
 * Some scanners are compiled-in; others can be loaded at run-time.
 * Scanners can have multiple feature recorders. Multiple scanners can record to the
 * same feature recorder.
 */

/* An array of the built-in scanners */
scanner_t *scanners_builtin[] = {
    scan_accts,
    scan_base64,
    scan_kml,
    scan_email,
    scan_gps,
    scan_net,
    scan_find,
    scan_wordlist,
    scan_aes,
    scan_json,
#ifdef HAVE_EXIV2
    scan_exif,
#endif
#ifdef HAVE_BE_EXIF
    scan_be_exif,
#endif
    scan_zip,
    scan_gzip,
    scan_pdf,
    scan_hiberfile,
    scan_winprefetch,
    0};


/****************************************************************
 *** COMMAND LINE OPTIONS
 ****************************************************************/

const char *progname=0;

size_t opt_pagesize=1024*1024*16;	// 
size_t opt_margin = 1024*1024*1;
int opt_notify_rate = 4;
int word_min = 6;
int word_max = 14;
int num_threads = 1;
int64_t opt_offset_start = 0;
int64_t opt_offset_end   = 0;
int min_uncompr_size = 6;	// don't bother with objects smaller than this
int max_uncompr_size = 256*1024*1024; // don't decompress objects larger than this

/****************************************************************
 *** PATH PRINTER
 ****************************************************************/

/* Get the next token from the path. Tokens are separated by dashes.*/
static string get_and_remove_token(string &path)
{
    while(path[0]=='-'){
	path = path.substr(1); // remove any leading dashes.
    }
    size_t dash = path.find('-');	// find next dash
    if(dash==string::npos){		// no string; return the rest
	string prefix = path;
	path = "";
	return prefix;
    }
    string prefix = path.substr(0,dash);
    path = path.substr(dash+1);
    return prefix;
}

void process_path_printer(const scanner_params &sp)
{
    /* 1. Get next token 
     * 2. if prefix part is a number, skip forward that much in sbuf and repeat.
     *    if the prefix is PRINT, print the buffer
     *    if next part is a string, strip it and run that decoder.
     *    if next part is a |, print
     */
    if(debug&1) cerr << "process_path_printer " << sp.sbuf.pos0.path << "\n";
    string new_path = sp.sbuf.pos0.path;
    string prefix = get_and_remove_token(new_path);

    /* Time to print ?*/
    if(prefix.size()==0 || prefix=="PRINT"){

	uint64_t print_start = 0;
	uint64_t print_len = 4096;
    
	/* Check for options */
	scanner_params::PrintOptions::iterator it;

	it = sp.print_options.find("Content-Length");
	if(it!=sp.print_options.end()){
	    print_len = pos0_t::stoi64(it->second);
	}

	it = sp.print_options.find("Range");
	if(it!=sp.print_options.end()){
	    if(it->second[5]=='='){
		size_t dash = it->second.find('-');
		string v1 = it->second.substr(6,dash-6);
		string v2 = it->second.substr(dash+1);
		print_start = pos0_t::stoi64(v1);
		print_len = pos0_t::stoi64(v2)-print_start+1;
	    }
	}

	if(print_start>sp.sbuf.bufsize){
	    print_len = 0;			// can't print anything
	}

	if(print_len>0 && print_start+print_len>sp.sbuf.bufsize){
	    print_len = sp.sbuf.bufsize-print_start;
	}

	switch(scanner_params::getPrintMode(sp.print_options)){
	case scanner_params::MODE_HTTP:
	    cout << "Content-Length: "		<< print_len  << HTTP_EOL;
	    cout << "Content-Range: bytes "	<< print_start << "-" << print_start+print_len-1 << HTTP_EOL;
	    cout << "X-Range-Available: bytes " << 0 << "-" << sp.sbuf.bufsize-1 << HTTP_EOL;
	    cout << HTTP_EOL;
	    sp.sbuf.raw_dump(cout,print_start,print_len); // send to stdout as binary
	    break;
	case scanner_params::MODE_RAW:
	    cout << print_len << HTTP_EOL;
	    cout.flush();
	    sp.sbuf.raw_dump(cout,print_start,print_len); // send to stdout as binary
	    break;
	case scanner_params::MODE_HEX:
	    sp.sbuf.hex_dump(cout,print_start,print_len);
	    break;
	case scanner_params::MODE_NONE:
	    break;
	}
	return;			// our job is done
    }
    /* If we are in an offset block, process recursively with the offset */
    if(isdigit(prefix[0])){
	uint64_t offset = pos0_t::stoi64(prefix);
	if(offset>sp.sbuf.bufsize){
	    printf("Error: %s only has %u bytes; can't offset to %u\n",
		   new_path.c_str(),(unsigned int)sp.sbuf.bufsize,(unsigned int)offset);
	    return;
	}
	process_path_printer(scanner_params(sbuf_t(new_path,
						   sp.sbuf.buf+offset,
						   sp.sbuf.bufsize-offset,
						   sp.sbuf.bufsize-offset),
					    sp.fs,sp.print_options));
	return;
    }
    /* Find the scanner and use it */
    for(scanner_vector::const_iterator it = current_scanners.begin();it!=current_scanners.end();it++){
	const string &name = upperstr((*it)->info.name);
	if(name==prefix){
	    ((*it)->scanner)(scanner_params(sbuf_t(new_path,sp.sbuf.buf,sp.sbuf.bufsize,sp.sbuf.bufsize),
					    sp.fs,sp.print_options),
			     recursion_control_block(process_path_printer,name,true));
	    return;
	}
    }
    cout << "Unknown name in path: " << prefix << "\n";
}

/**
 * process the image for a given path.
 */


static void process_path(const image_process &p,string path,scanner_params::PrintOptions &po)
{
    /* Check for "/r" in path which means print raw */
    if(path.size()>2 && path.substr(path.size()-2,2)=="/r"){
	path = path.substr(0,path.size()-2);
    }

    string prefix = get_and_remove_token(path);
    int64_t offset = pos0_t::stoi64(prefix);

    /* Get the offset. process */
    u_char *buf = (u_char *)calloc(opt_pagesize,1);
    if(!buf) errx(1,"Cannot allocate buffer");
    int count = p.pread(buf,opt_pagesize,offset);
    if(count<0){
	cerr << p.image_fname << ": " << strerror(errno) << " (Read Error)\n";
	return;
    }

    /* make up a bogus feature recorder set and feature recorder */
    feature_recorder_set fs(feature_recorder_set::DISABLED);

    pos0_t pos0(path+"-PRINT"); // insert the PRINT token
    sbuf_t sbuf(pos0,buf,count,count);
    scanner_params sp(sbuf,fs,po);
    process_path_printer(sp);
}

/**
 * process a path for a given filename.
 * Opens the image and calls the function above.
 * Also implements HTTP server with "-http" option.
 */
static void process_path(const char *fn,string path)
{
    image_process *pp = image_process_open(fn,0);

    if(path=="-"){
	/* process path interactively */
	printf("Path Interactive Mode:\n");
	if(pp==0){
	    cerr << "Invalid file name: " << fn << "\n";
	    exit(1);
	}
	do {
	    getline(cin,path);
	    if(path==".") break;
	    scanner_params::PrintOptions po;
	    scanner_params::setPrintMode(po,scanner_params::MODE_HEX);
	    process_path(*pp,path,po);
	} while(true);
	return;
    }
    if(path=="-http"){
	if(pp==0){
	    cout << "HTTP/1.1 502 Filename " << fn << " is invalid" << HTTP_EOL << HTTP_EOL;
	    exit(1);
	}
	do {
	    /* get the HTTP query */
	    string line;		// the specific query
	    scanner_params::PrintOptions po;
	    scanner_params::setPrintMode(po,scanner_params::MODE_HTTP);	// options for this query
	    
	    getline(cin,line);
	    truncate_at(line,'\r');
	    if(line.substr(0,4)!="GET "){
		cout << "HTTP/1.1 501 Method not implemented" << HTTP_EOL << HTTP_EOL;
		return;
	    }
	    size_t space = line.find(" HTTP/1.1");
	    if(space==string::npos){
		cout << "HTTP/1.1 501 Only HTTP/1.1 is implemented" << HTTP_EOL << HTTP_EOL;
		return;
	    }
	    string p2 = line.substr(4,space-4);

	    /* Get the additional header options */
	    do {
		getline(cin,line);
		truncate_at(line,'\r');
		if(line.size()==0) break; // double new-line
		size_t colon = line.find(":");
		if(colon==string::npos){
		    cout << "HTTP/1.1 502 Malformed HTTP request" << HTTP_EOL;
		    return;
		}
		string name = line.substr(0,colon);
		string val  = line.substr(colon+1);
		while(val.size()>0 && (val[0]==' '||val[0]=='\t')) val = val.substr(1);
		po[name]=val;
	    } while(true);
	    /* Process some specific URLs */
	    if(p2=="/info"){
		cout << "X-Image-Size: " << pp->image_size() << HTTP_EOL;
		cout << "X-Image-Filename: " << pp->image_fname << HTTP_EOL;
		cout << "Content-Length: 0" << HTTP_EOL;
		cout << HTTP_EOL;
		continue;
	    }

	    /* Ready to go with path and options */
	    process_path(*pp,p2,po);
	} while(true);
	return;
    }
    scanner_params::PrintOptions po;
    scanner_params::print_mode_t mode = scanner_params::MODE_HEX;
    if(path.size()>2 && path.substr(path.size()-2,2)=="/r"){
	path = path.substr(0,path.size()-2);
	mode = scanner_params::MODE_RAW;
    }
    if(path.size()>2 && path.substr(path.size()-2,2)=="/h"){
	path = path.substr(0,path.size()-2);
	mode = scanner_params::MODE_HEX;
    }
    scanner_params::setPrintMode(po,mode);
    process_path(*pp,path,po);
}

/****************************************************************
 *** Phase 1 BUFFER PROCESSING
 ****************************************************************/


/** process_extract is the main workhorse. It is calls each scanner on each page. 
 * @param sp    - the scanner params, including the sbuf to process 
 * It is also the recursive entry point for sub-analysis.
 */

void process_extract(const class scanner_params &sp)
{
    const sbuf_t &sbuf = sp.sbuf;
    const pos0_t &pos0 = sp.sbuf.pos0;
    class feature_recorder_set &fs = sp.fs;

    if(debug & DEBUG_NO_SCANNERS) {
	cerr << "DEBUG_NO_SCANNERS sbuf:" << sbuf << "\n";
	return;
    }
    if(debug & DEBUG_PRINT_STEPS) {
	cerr << "process_extract(" << pos0 << "," << sbuf << ")\n";
    }
    if(debug & DEBUG_DUMP_DATA) {
	sbuf.hex_dump(cerr);
    }

    if(sp.depth >= scanner_def::max_depth){
	feature_recorder_set::alert_recorder->write(pos0,"process_extract: MAX DEPTH REACHED","");
	return;
    }

    /****************************************************************
     *** CALL EACH OF THE SCANNERS ON THE SBUF
     ****************************************************************/

    for(scanner_vector::iterator it = current_scanners.begin();it!=current_scanners.end();it++){
	if((*it)->enabled==false) continue;
	string name = (*it)->info.name;
	
	try {
	    if(debug & DEBUG_PRINT_STEPS){
		cerr << "calling scanner " << name << "...\n";
	    }
		    
	    /* Compute the effective path for stats */
	    bool inname=false;
	    string epath;
	    for(string::const_iterator cc=sp.sbuf.pos0.path.begin();cc!=sp.sbuf.pos0.path.end();cc++){
		if(isupper(*cc)) inname=true;
		if(inname) epath.push_back(toupper(*cc));
		if(*cc=='-') inname=false;
	    }
	    if(epath.size()>0) epath.push_back('-');
	    for(string::const_iterator cc=name.begin();cc!=name.end();cc++){
		epath.push_back(toupper(*cc));
	    }

	    /* Call the scanner.*/
	    aftimer t;
	    t.start();
	    recursion_control_block rcb(process_extract,upperstr(name),false);
	    ((*it)->scanner)(sp,rcb);
	    t.stop();
	    sp.fs.add_stats(epath,t.elapsed_seconds());
		    
	    if(debug & DEBUG_PRINT_STEPS){
		cerr << "     ... scanner " << name << " returned\n";
	    }
	}
	catch ( exception &e ) {
	    cerr << "Scanner  " << name << " Exception: " << e.what() << "\n";
	}
    }
    fs.flush_all();
}


/**
 * Sleep for a minimum of msec
 */
static void msleep(int msec)
{
#if _WIN32
    Sleep(msec);
    return;
#else
#  ifdef HAVE_USLEEP
    usleep(msec*1000);
#  else
    int sec = msec/1000;
    if(sec<1) sec=1;
    sleep(sec);			// posix
#  endif
#endif
}

/**
 * phase1 - For every page of the iterator, schedule work.
 */

static void phase1(image_process &p,feature_recorder_set &fs,
		   xml &xreport,
		   int64_t &page_number,int64_t &total_bytes,
		   xml::tagid_set_t &seen_page_ids,
		   aftimer &timer)
{
    int zero_pages = 0;
    md5_generator *md5g = new md5_generator();		// keep track of MD5
    uint64_t md5_next = 0;		// next byte to hash
    threadpool tp(num_threads,fs,xreport);
    
    xreport.push("runtime","xmlns:debug=\"http://www.afflib.org/bulk_extractor/debug\"");
    cout << "Phase 1.\n";

    for(image_process::iterator it = p.begin(); it!= p.end() ; ++it){
	if(opt_offset_end!=0 && it.raw_offset > opt_offset_end) break; // passed the offset
	if(page_number>=opt_page_start && it.raw_offset>=opt_offset_start){
	    if(seen_page_ids.find(it.get_pos0().str()) != seen_page_ids.end()){
		// We've seen this page; skip it.
		continue;
	    }

	    // attempt to get an sbuf. If we can't get it, we may be in a low-memory situation.
	    // wait for 30 seconds.

	    sbuf_t *sbuf = 0;
	    for(int retry_count=0;retry_count<max_sbuf_alloc_errors && sbuf==0;retry_count++){
		try {
		    sbuf = it.sbuf_alloc();
		    if(sbuf==0){
			cerr << "sbuf_alloc() failed. Will retry...\n";
		    }
		}
		catch (const std::exception &e) {
		    cerr << "sbuf_alloc failed: " << e.what() << " (retry=" << retry_count << ")\n";
		    cerr << "will wait for 60 seconds and try again...\n";
		    msleep(60000);
		}
		if(retry_count+1>=max_sbuf_alloc_errors){
		    cerr << "Too many errors encountered in a row. Diagnose and restart.\n";
		    exit(1);
		}
	    }
	    sbuf->page_number = page_number;
	    if(sbuf->pagesize==0){
		cerr << "Page " << page_number << " has no data\n";
		zero_pages++;
		if(zero_pages>max_zero_pages){
		    cout << "Zero-length pages read: " << zero_pages <<  "zero-length pages found.";
		    //break;
		}
	    }
	    if(md5g){
		if(sbuf->pos0.offset==md5_next){ 
		    // next byte follows logically
		    md5g->update(sbuf->buf,sbuf->pagesize);
		    md5_next += sbuf->pagesize;
		} else {
		    // we had a logical gap; stop hashing
		    delete md5g;	
		    md5g = 0;
		}
	    }
		
	    /* Notify the user */
	    if((opt_quiet==0) || (sbuf->page_number % opt_quiet)==0){
		time_t t = time(0);
		struct tm tm;
		localtime_r(&t,&tm);
		printf("%2d:%02d:%02d %s (%4.2f%%) Done in %s at %s\n",
		       tm.tm_hour,tm.tm_min,tm.tm_sec,
		       it.str().c_str(),
		       it.fraction_done()*100.0,
		       timer.eta_text(it.fraction_done()).c_str(),
		       timer.eta_time(it.fraction_done()).c_str());
	    }
	    total_bytes += sbuf->pagesize;
	    tp.schedule_work(sbuf);	// schedule the work
	}
	page_number++;
    }
	    
    cout << "All Data is Read; waiting for threads to finish...\n\n";

    /* Now wait for all of the threads to be free */
    time_t wait_start = time(0);
    const time_t max_wait_time=3600;
    while (true){
	msleep(250);
	int time_waiting  = time(0) - wait_start;
	int num_remaining = num_threads - tp.get_free_count();
	int sec_waiting = time_waiting % 60;
	int min_waiting = time_waiting / 60;
	if(num_remaining==0) break;

	
	std::stringstream ss;

	ss << "Time elapsed waiting for " << num_remaining
	   << " thread" << (num_remaining>1 ? "s" : "") 
	   << " to finish: ";
	if(min_waiting) ss << min_waiting << " min ";
	if(sec_waiting) ss << sec_waiting << " sec ";
	ss << " (please wait for another " << (max_wait_time - time_waiting) << " seconds)";
	
	if(opt_quiet==0)   cout << "\r" << ss.str() << "\r";
	if(sec_waiting==0) xreport.xmlcomment(ss.str());

	if(time_waiting>max_wait_time){
	    cout << "\n\n";
	    cout << " ... this shouldn't take more than an hour. Exiting ... \n";
	    cout << " ... Please report to the bulk_extractor maintainer ... \n";
	    break;
	}
    }
    cout << "All Threads Finished!\n";
    xreport.pop();			// pop runtime
    /* We can write out the source info now, since we (might) know the hash */
    xreport.push("source");
    xreport.xmlout("image_filename",p.image_fname);
    xreport.xmlout("image_size",p.image_size());  
    if(md5g){
	md5_t md5 = md5g->final();
	xreport.xmlout("hashdigest",md5.hexdigest(),"type='MD5'",false);
	delete md5g;
    }
    xreport.pop();			// source
    /* Record the feature files and their counts in the output */
    xreport.push("feature_files");
    for(feature_recorder_map::const_iterator it = tp.fs.frm.begin(); it != tp.fs.frm.end(); it++){
	xreport.push("feature_file");
	xreport.xmlout("name",it->second->name);
	xreport.xmlout("count",it->second->count);
	xreport.pop();
    }
    xreport.pop();
    
}

/****************************************************************
 *** PHASE 2: shut down the scanners
 ****************************************************************/

void phase2(feature_recorder_set &fs)
{
    cout << "Phase 2. Shutting down scanners\n";
    for(scanner_vector::iterator it = current_scanners.begin();it!=current_scanners.end();it++){
	if((*it)->enabled){
	    pos0_t pos0;
	    sbuf_t sbuf(pos0,0,0,0);
	    scanner_params sp(sbuf,fs);
	    recursion_control_block rcb(0,"",0);	// empty rcb 
	    sp.phase=2;				// shutdown
	    (*(*it)->scanner)(sp,rcb);
	}
    }
}

/****************************************************************
 *** PHASE 3: Create the histograms
 ****************************************************************/
static void phase3(feature_recorder_set &fs, xml &xreport)
{
    cout << "Phase 3. Creating Histograms\n";
    for(histograms_t::const_iterator it = histograms.begin();it!=histograms.end();it++){
	if(fs.has_name((*it).feature)){
	    feature_recorder *fr = fs.get_name((*it).feature);
	    try {
		fr->make_histogram((*it).pattern,(*it).suffix);
	    }
	    catch (const std::exception &e) {
		cerr << "ERROR: " << e.what() << " computing histogram " << (*it).feature << "\n";
		xreport.xmlout("error",(*it).feature, "function='phase3'",true);
	    }
	}
    }
}

/****************************************************************
 *** Usage
 ****************************************************************/

static void usage() __attribute__ ((__noreturn__));
static void usage()
{
    cerr << "bulk_extractor version " PACKAGE_VERSION " " << svn_revision << "\n";
    cerr << "Usage: " << progname << " [options] imagefile\n";
    cerr << "  runs bulk extractor and outputs to stdout a summary of what was found where\n";
    cerr << "\n";
    cerr << "Required parameters:\n";
    cerr << "   imagefile     - the file to extract\n";
#ifdef HAVE_FTS_OPEN
    cerr << " or  -R filedir  - recurse through a directory of files\n";
#endif
#ifdef HAVE_LIBEWF
    cerr << "                  SUPPORT FOR E01 FILES COMPILED IN\n";
#endif
#ifdef HAVE_LIBAFFLIB
    cerr << "                  SUPPORT FOR AFF FILES COMPILED IN\n";
#endif    
#ifdef HAVE_EXIV2
    cerr << "                  EXIV2 COMPILED IN\n";
#endif    
#ifdef HAVE_BE_EXIF
    cerr << "                  BE_EXIF COMPILED IN\n";
#endif    
    cerr << "   -o outdir    - specifies output directory. Must not exist.\n";
    cerr << "                  bulk_extractor creates this directory.\n";
    cerr << "Options:\n";
    cerr << "   -b banner.txt- Add banner.txt contents to the top and bottom of every output file.\n";
    cerr << "   -r alert_list.txt  - a file containing the alert list of features to alert\n";
    cerr << "                       (can be a feature file or a list of globs)\n";
    cerr << "                       (can be repeated.)\n";
    cerr << "   -w stop_list.txt   - a file containing the stop list of features (white list\n";
    cerr << "                       (can be a feature file or a list of globs)s\n";
    cerr << "                       (can be repeated.)\n";
    cerr << "   -F <rfile>   - Read a list of regular expressions from <rfile> to find\n";
    cerr << "   -f <regex>   - find occurances of <regex>; may be repeated.\n";
    cerr << "                  results go into find.txt\n";
    cerr << "   -q nn        - Quiet; only print every nn status reports\n";
    cerr << "\nTuning parameters:\n";
    cerr << "   -C NN         - specifies the size of the context window (default " << feature_recorder::context_window << ")\n";
    cerr << "   -G NN         - specify the page size (default " << opt_pagesize << ")\n";
    cerr << "   -g NN         - specify margin (default " <<opt_margin << ")\n";
    cerr << "   -W n1:n2      - Specifies minimum and maximum word size\n";
    cerr << "                  (default is -w" << word_min << ":" << word_max << "\n";
    cerr << "   -j NN         - Number of threads to run (default " <<threadpool::numCPU() << ")\n";
    cerr << "\nPath Processing Mode:\n";
    cerr << "   -p <path>/f  - print the value of <path> with a given format.\n";
    cerr << "                  formats: r = raw; h = hex.\n";
    cerr << "                  Specify -p - for interactive mode.\n";
    cerr << "                  Specify -p -http for HTTP mode.\n";
    cerr << "\nParallelizing:\n";
    cerr << "   -Y <o1>      - Start processing at o1 (o1 may be 1, 1K, 1M or 1G)\n";
    cerr << "   -Y <o1>-<o2> - Process o1-o2\n";
    cerr << "   -A <off>     - Add <off> to all reported feature offsets\n";
    cerr << "\nDebugging:\n";
    cerr << "   -V print version number\n";
    cerr << "   -c           - Enable Crash Protection\n";
    cerr << "   -M nn        - sets max recursion depth (default " << scanner_def::max_depth << ")\n";
    cerr << "   -z nn        - start on page nn\n";
    cerr << "   -dN          - debug mode (see source code\n";
    cerr << "   -Z           - zap (erase) output directory\n";
    cerr << "\nControl of Scanners:\n";
    cerr << "   -P <dir>     - Specifies a plugin directory\n";
    cerr << "   -E scanner   - turn off all scanners except scanner\n";
    cerr << "   -m <max>     - maximum number of minutes to wait for memory starvation\n";
    cerr << "                  default is " << max_sbuf_alloc_errors << "\n";

    /* Print a list of scanners */
    load_scanners(scanners_builtin,histograms);
    cerr << "\n";
    for(scanner_vector::const_iterator it = current_scanners.begin();it!=current_scanners.end();it++){
	if((*it)->info.flags & scanner_info::SCANNER_NO_USAGE) continue;
	if((*it)->info.flags & scanner_info::SCANNER_DISABLED){
	    cerr << "   -e " << (*it)->info.name << " - enable scanner " << (*it)->info.name << "\n";
	}
    }
    cerr << "\n";
    for(scanner_vector::const_iterator it = current_scanners.begin();it!=current_scanners.end();it++){
	if((*it)->info.flags & scanner_info::SCANNER_NO_USAGE) continue;
	if(((*it)->info.flags & scanner_info::SCANNER_DISABLED)==0){
	    cerr << "   -x " << (*it)->info.name << " - disable scanner " << (*it)->info.name << "\n";
	}
    }
#if !(defined HAVE_BE_EXIF || defined HAVE_EXIV2)
    cerr << "WARNING: exif carving not compiled in\n";
#endif    
    exit(1);
}


/*
 * Make sure that the filename provided is sane.
 * That is, do not allow the analysis of a *.E02 file...
 */
void validate_fn(const string &fn)
{
    if(access(fn.c_str(),R_OK)){
	cerr << "file not found: " << fn << "\n";
	exit(1);
    }
    if(fn.size()>3){
	size_t e = fn.rfind('.');
	if(e!=string::npos){
	    string ext = fn.substr(e+1);
	    if(ext=="E02" || ext=="e02"){
		cerr << "Error: invalid file name\n";
		cerr << "Do not use bulk_extractor to process individual EnCase files.\n";
		cerr << "Instead, just run bulk_extractor with FILENAME.E01\n";
		cerr << "The other files in an EnCase multi-volume archive will be opened\n";
		cerr << "automatically.\n";
		exit(1);
	    }
	}
    }
}

/**
 * Create the dfxml output
 */

static void dfxml_create(xml &xreport,const string &command_line)
{
    xreport.push("dfxml","xmloutputversion='1.0'");
    xreport.push("metadata",
		 "\n  xmlns='http://afflib.org/bulk_extractor/' "
		 "\n  xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' "
		 "\n  xmlns:dc='http://purl.org/dc/elements/1.1/'" );
    xreport.xmlout("dc:type","Feature Extraction","",false);
    xreport.pop();
    xreport.add_DFXML_creator(PACKAGE_NAME,PACKAGE_VERSION,svn_revision_clean(),command_line);

    xreport.push("configuration");
    xreport.xmlout("threads",num_threads);
    xreport.push("scanners");
    /* Generate a list of the scanners in use */

    for(scanner_vector::const_iterator it=current_scanners.begin();it!=current_scanners.end();it++){
	if((*it)->enabled){
	    xreport.xmlout("scanner",(*it)->info.name);
	}
    }
    xreport.pop();			// scanners
    xreport.pop();			// configuration
}



int main(int argc,char **argv)
{
    assert(sizeof(off_t)==8);		// be sure that we were compiled correctly
    progname = argv[0];
    int64_t page_number = 0;
    //const char *opt_context_stoplist = "";
    const char *opt_path = 0;
    int opt_recurse = 0;
    int opt_zap = 0;
    string opt_outdir;
    char *cc;
    setvbuf(stdout,0,_IONBF,0);		// don't buffer stdout
    string command_line = xml::make_command_line(argc,argv);
    num_threads = threadpool::numCPU();

#ifdef WIN32
    setmode(1,O_BINARY);		// make stdout binary
    threadpool::win32_init();
#endif
    /* look for usage first */
    if(argc==1 || (strcmp(argv[1],"-h")==0)){
	usage();
	exit(1);
    }

    /* Load the scanners */
    load_scanners(scanners_builtin,histograms);		  // and load the scanners

    /* Process options */
    int ch;
    while ((ch = getopt(argc, argv, "A:b:C:d:E:e:F:f:G:g:hM:m:o:P:p:q:Rr:j:VW:w:x:Y:z:Z")) != -1) {
	switch (ch) {
	case 'A': feature_recorder::offset_add  = pos0_t::stoi64(optarg);break;
	case 'b': feature_recorder::banner_file = optarg; break;
	case 'C': feature_recorder::context_window = atoi(optarg);break;
	case 'd':
	    {
		int d = atoi(optarg);
		switch(d){
		case 1002: 
		    if(calloc(1024*1024*512,1)){
		      cerr << "-d1002 -- Allocating 512MB of RAM; may be repeated\n";
		    } else {
		      cerr << "-d1002 -- CANNOT ALLOCATE MORE RAM\n";
		    }
		    break;
		case 1003:
		    {
			xml x("test.xml",false);
			int line = 0;
			while(1){
			    x.xmlout("line",line++);
			    std::cout << "line="<<line<<"\n";
			    msleep(500);
			}
		    }
		default:
		    debug  = d;
		    break;
		}
	    }
	    break;
	case 'E':
	    disable_all_scanners();
	    set_scanner_enabled(optarg,1);
	    break;
	case 'e': set_scanner_enabled(optarg,1);break;
	case 'F': process_find_file(optarg); break;
	case 'f': add_find_pattern(optarg); break;
	case 'G': opt_pagesize = atoi(optarg); break;
	case 'g': opt_margin = atoi(optarg); break;
	case 'j': num_threads = atoi(optarg); break;
	case 'M': scanner_def::max_depth = atoi(optarg); break;
	case 'm': max_sbuf_alloc_errors = atoi(optarg); break;
	case 'o': opt_outdir = optarg;break;
	case 'P': load_scanner_directory(optarg,histograms);break;
	case 'p': opt_path = optarg; break;
        case 'q': opt_quiet = atoi(optarg);break;
	case 'r': if(alert_list.readfile(optarg)){
		err(1,"Cannot read alert list %s",optarg);
	    }
	    break;
	case 'R': opt_recurse = 1; break;
	case 'V': cout << "bulk_extractor " << PACKAGE_VERSION << "\n"; exit (1);
	case 'W':
	    cc = strchr(optarg,':');
	    if(!cc) err(1,"-W requires n1:n2");
	    cc[0] = 0;			// null terminate
	    word_min = atoi(optarg);
	    word_max = atoi(cc+1);
	    if(word_min>word_max) err(1,"word_min=%d word_max=%d\n",word_min,word_max);
	    break;
	case 'w': if(stop_list.readfile(optarg)){
		err(1,"Cannot read stop list %s",optarg);
	    }
	    break;
	case 'x': set_scanner_enabled(optarg,0);break;
	case 'Y': {
	    string optargs = optarg;
	    size_t dash = optargs.find('-');
	    if(dash==string::npos){
		opt_offset_start = pos0_t::stoi64(optargs);
	    } else {
		opt_offset_start = pos0_t::stoi64(optargs.substr(0,dash-1));
		opt_offset_end   = pos0_t::stoi64(optargs.substr(dash+1));
	    }
	    break;
	}
	    
	case 'z': opt_page_start = pos0_t::stoi64(optarg);break;
	case 'Z': opt_zap=true;break;
	case 'h': case '?':default:
	    usage();
	    break;
	}
    }
    argc -= optind;
    argv += optind;

    if(opt_path){
	if(argc!=1) errx(1,"-p requires a single argument.");
	process_path(argv[0],opt_path);
	exit(0);
    }
    if(opt_outdir.size()==0) err(1,"error: -o outdir must be specified");

    /* The zip option wipes the contents of a directory, useful for debugging */
    if(opt_zap){
	DIR *dirp = opendir(opt_outdir.c_str());
	if(dirp){
	    struct dirent *dp;
	    while ((dp = readdir(dirp)) != NULL){
		string name = dp->d_name;
		if(name=="." || name=="..") continue;
		string fname = opt_outdir + string("/") + name;
		unlink(fname.c_str());
		cout << "erasing " << fname << "\n";
	    }
	}
	if(rmdir(opt_outdir.c_str())){
	    cerr << "cannot rmdir " << opt_outdir << "\n";
	    exit(1);
	}
	cout << "rmdir " << opt_outdir << "\n";
    }

    /* Start the clock */
    aftimer timer;
    timer.start();

    /* If output directory does not exist, we are not restarting! */
    xml  *xreport=0;
    string reportfilename = opt_outdir + "/report.xml";

    xml::tagid_set_t seen_page_ids; // pages that do not need re-processing
    image_process *p = 0;

    if(access(opt_outdir.c_str(),F_OK)!=0){
	/* Create the new directory */
	if ( argc !=1 ) errx(1,"Disk image option not provided. Run with -h for help.");

	/* If disk image does not directory exists, we are in restart mode */
	validate_fn(*argv);
	be_mkdir(opt_outdir);
	p = image_process_open(*argv,opt_recurse);

	/* Store the configuration in the XML file */
	xreport = new xml(reportfilename,false);
	dfxml_create(*xreport,command_line);
	xreport->xmlout("provided_filename",*argv); // save this information
    } else {
	/* Restarting */
	if(access(reportfilename.c_str(),R_OK)){
	    cerr << opt_outdir << ": error\n";
	    cerr << "report.xml file is missing or unreadable.\n";
	    cerr << "Directory may not have been created by bulk_extractor.\n";
	    cerr << "Cannot continue.\n";
	    exit(1);
	}
	
	/* Specify the tags for which we are searching for */
	xml::tagmap_t tagmap;
	tagmap["provided_filename"] = "";
	string dws("debug:work_start");
	string attrib = "pos0";

	class xml::existing e;
	e.tagmap = &tagmap;
	e.tagid = &dws;
	e.attrib = &attrib;
	e.tagid_set = &seen_page_ids;

	xreport = new xml(reportfilename,e);

	if(tagmap["provided_filename"].size()==0){
	    cerr << "report.xml in output directory is incomplete.\n";
	    cerr << "cannot restart.\n";
	    exit(1);
	}
	
	if(tagmap["provided_filename"] != *argv){
	    cerr << "report.xml in output directory is for " << tagmap["provided_filename"] << "\n";
	    cerr << "cannot restart with " << *argv << "\n";
	    exit(1);
	}

	p = image_process_open(*argv,opt_recurse);
	cout << "Restarting from " << opt_outdir << "\n";
    }

    if(p==0) err(1,"Cannot image_process_open(%s)",*argv);
	    
    /* SLG --- TK --- just mak feature_recorder_set take a list of the scanners? */
    /* Determine the feature files that will be used */
    feature_file_names_t feature_file_names;			// current feature files
    feature_file_names.insert(feature_recorder_set::ALERT_RECORDER); // we always have alerts
    for(scanner_vector::const_iterator it=current_scanners.begin();it!=current_scanners.end();it++){
	if((*it)->enabled){
	    for(set<string>::const_iterator fi=(*it)->info.feature_names.begin();
		fi!=(*it)->info.feature_names.end();
		fi++){
		feature_file_names.insert(*fi);
	    }
	}
    }
    feature_recorder_set fs(feature_file_names,opt_outdir);
    feature_recorder_set::alert_recorder = fs.get_name(feature_recorder_set::ALERT_RECORDER);
    int64_t total_bytes = 0;

    /* provide documentation to the user; the DFXML information comes from elsewhere */
    cout << "bulk_extractor version:" << PACKAGE_VERSION << "\n";
#ifdef HAVE_GETHOSTNAME
    char hostname[1024];
    gethostname(hostname,sizeof(hostname));
    cout << "Hostname: " << hostname << "\n";
#endif
    cout << "Input file: " << *argv << "\n";
    cout << "Output directory: " << opt_outdir << "\n";
    cout << "Disk Size: " << p->image_size() << "\n";
    cout << "Threads: " << num_threads << "\n";

    /****************************************************************
     *** THIS IS IT! 
     ****************************************************************/

    phase1(*p,fs,*xreport,page_number,total_bytes,seen_page_ids,timer);
    phase2(fs);
    phase3(fs,*xreport);

    /* Print usage information */
    printf("# elapsed time: %.1f seconds\n",timer.elapsed_seconds());
    xreport->push("report");
    xreport->xmlout("total_bytes",total_bytes);
    xreport->xmlout("elapsed_seconds",timer.elapsed_seconds());
    xreport->pop();			// report
    fs.dump_stats(*xreport);
    xreport->add_rusage();
    xreport->pop();			// bulk_extractor
    xreport->close();
    delete p;				// not strictly needed, but why not?
    exit(0);
}

