#ifndef BULK_EXTRACTOR_I_H
#define BULK_EXTRACTOR_I_H

/**
 * bulk_extractor scanner plug_in architecture.
 * 
 * Scanners are called with two parameters:
 * A reference to a scanner_params (SP) object.
 * A reference to a recursion control block (RCB) object.
 * 
 * On startup, each scanner is called with a special SP and RCB.
 * The scanners respond by setting fields in the SP and returning.
 * 
 * When executing, once again each scanner is called with the SP and RCB.
 * By design, this file can be read without reading config.h
 * This is the only file that needs to be included for a scanner.
 *
 * phase0 - scanners are loaded and register the names of the feature files they want.
 * phase1 - each scanner is called to analyze 1 or more sbufs.
 * phase2 - scanners are given a chance to shutdown
 */

#ifndef	__cplusplus
#error bulk_extractor_i.h requires C++
#endif

#include "sbuf.h"
#include "cppmutex.h"
#include "feature_recorder.h"
#include "feature_recorder_set.h"

/**
 * The scanner params class is the primary way that the bulk_extractor framework
 * communicates with the scanners. 
 * @param sbuf - the buffer to be scanned
 * @param feature_names - if fs==0, add to feature_names the feature file types that this
 *                        scanner records.. The names can have a /c appended to indicate
 *                        that the feature files should have context enabled. Do not scan.
 * @param fs   - where the features should be saved. Must be provided if feature_names==0.
 **/

class histogram_def {
 public:
 histogram_def(string feature_,string pattern_,string suffix_):
    feature(feature_),pattern(pattern_),suffix(suffix_){}
    string feature;			/* feature file */
    string pattern;			/* extract pattern; "" means use entire feature */
    string suffix;			/* suffix to append; "" means "histogram" */
};

typedef  set<histogram_def> histograms_t;

inline bool operator <(class histogram_def h1,class histogram_def h2)  {
    if (h1.feature<h2.feature) return true;
    if (h1.feature>h2.feature) return false;
    if (h1.pattern<h2.pattern) return true;
    if (h1.pattern>h2.pattern) return false;
    if (h1.suffix<h2.suffix) return true;
    if (h1.suffix>h2.suffix) return false;
    return false;			/* equal */
};


inline bool operator !=(class histogram_def h1,class histogram_def h2)  {
    return h1.feature!=h2.feature || h1.pattern!=h2.pattern || h1.suffix!=h2.suffix;
};

typedef void scanner_t(const class scanner_params &sp,const class recursion_control_block &rcb);
typedef void process_t(const class scanner_params &sp); 

/** scanner_info gets filled in by the scanner to tell the caller about the scanner.
 */
class scanner_info {
 public:
    static const int SCANNER_DISABLED=0x01;		/* v1: enabled by default */
    static const int SCANNER_NO_USAGE=0x02;		/* v1: do not show scanner in usage */

    scanner_info():si_version(0),name(),author(),description(),url(),scanner_version(),
		   flags(0),feature_names(),
		    histogram_defs(){}
    int		si_version;		// version number for this structure
    string	name;			// v1: scanner name
    string	author;			// v1: who wrote me?
    string	description;		// v1: what do I do?
    string	url;			// v1: where I come from
    string	scanner_version;	// v1: version for the scanner
    uint64_t	flags;			// v1: flags
    set<string> feature_names;		// v1: features I need
    histograms_t histogram_defs;	// v1: histogram definition info
};

class scanner_params {
 public:
    /** Construct a scanner_params from a sbuf and other sensible defaults.
     *
     */
    enum print_mode_t {MODE_NONE=0,MODE_HEX,MODE_RAW,MODE_HTTP};
    typedef tr1::unordered_map<string,string> PrintOptions;
    static print_mode_t getPrintMode(const PrintOptions &po){
	PrintOptions::const_iterator p = po.find("print_mode_t");
	if(p != po.end()){
	    if(p->second=="MODE_NONE") return MODE_NONE;
	    if(p->second=="MODE_HEX") return MODE_HEX;
	    if(p->second=="MODE_RAW") return MODE_RAW;
	    if(p->second=="MODE_HTTP") return MODE_HTTP;
	}
	return MODE_NONE;
    }
    static void setPrintMode(PrintOptions &po,int mode){
	switch(mode){
	default:
	case MODE_NONE:po["print_mode_t"]="MODE_NONE";return;
	case MODE_HEX:po["print_mode_t"]="MODE_HEX";return;
	case MODE_RAW:po["print_mode_t"]="MODE_RAW";return;
	case MODE_HTTP:po["print_mode_t"]="MODE_HTTP";return;
	}
    }

    /* A scanner params with all of the instance variables */
    scanner_params(const sbuf_t &sbuf_,class feature_recorder_set &fs_, PrintOptions &print_options_):
	version(1),phase(1),sbuf(sbuf_),fs(fs_),depth(0),
	print_options(print_options_), info(0){
    }

    /* A scanner params with no print options*/
    static PrintOptions no_options;	// in common.cpp
    scanner_params(const sbuf_t &sbuf_,class feature_recorder_set &fs_):
	version(1),phase(1),sbuf(sbuf_),fs(fs_),depth(0), print_options(no_options),info(0){
    }

    /** Construct a scanner_params for recursion from an existing sp and a new sbuf.
     */
    scanner_params(const scanner_params &sp_existing,const sbuf_t &sbuf_new):
	version(1),phase(1),
	sbuf(sbuf_new),fs(sp_existing.fs),depth(sp_existing.depth+1),
	print_options(sp_existing.print_options),info(0){
    }

    int   version;			/* version number */
    int	  phase;			/* v1: 0=startup, 1=normal, 2=shutdown */
    const sbuf_t &sbuf;			/* v1: what to scan */
    class feature_recorder_set &fs;	/* v1: where to put the results*/
    uint32_t   depth;			/* v1: how far down are we? */

    /* These are for printing */
    PrintOptions  &print_options;	/* v1: how to print */
    scanner_info  *info;		/* v1: get parameters on startup; info's are stored in a the scanner_def vector */
};


inline std::ostream & operator <<(std::ostream &os,const class scanner_params &sp){
    os << "scanner_params(" << sp.sbuf << ")";
    return os;
};

class recursion_control_block {
 public:
 recursion_control_block(process_t *callback_,string partName_,bool raf):
    callback(callback_),partName(partName_),returnAfterFound(raf){}
    process_t *callback;
    string partName;		/* eg "ZIP", "GZIP" */
    bool returnAfterFound;	/* only run once */
};
    
/* plugin.cpp */
class scanner_def {
public:;
    static uint32_t max_depth;
    scanner_def():scanner(0),enabled(false),info(),pathPrefix(){};
    const scanner_t  *scanner;		// pointer to the primary entry point
    bool	enabled;		// is enabled?
    scanner_info info;
    string	pathPrefix;		/* path prefix for recursive scanners */
};
void load_scanner(const scanner_t &scanner,histograms_t &histograms);
void load_scanners(const scanner_t *scanners[],histograms_t &histograms);		// load the scan_ plugins
void load_scanner_directory(const string &dirname,histograms_t &histograms);		// load the scan_ plugins
void disable_all_scanners();
typedef vector<scanner_def *> scanner_vector;
extern scanner_vector current_scanners;				// current scanners

#endif
