/**
 * exif_entry: - reads an entry of an IFD type to provide a name and value pair
 *               as strings for use by bulk_extractor
 * Revision history:
 * 2012-jan-05 bda - Created.
 *
 * Reading approach: read every entry available,
 * but disregard entries whose size conflicts with the tag.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <cassert>
#include <algorithm>
#include "../config.h"

#ifdef HAVE_BE_EXIF

#include "exif_reader.h"

// entry types
static const uint16_t EXIF_BYTE = 1;		// uint8
static const uint16_t EXIF_ASCII = 2;		// byte with 7-bit ascii
static const uint16_t EXIF_SHORT = 3;		// uint16
static const uint16_t EXIF_LONG = 4;		// uint32
static const uint16_t EXIF_RATIONAL = 5;	// two LONGs
static const uint16_t EXIF_UNDEFINED = 7;	// byte whose value depends on the field definition
static const uint16_t EXIF_SLONG = 9;		// signed int32
static const uint16_t EXIF_SRATIONAL = 10;	// two SLONGs

// private helpers
static void add_entry(ifd_type_t ifd_type, const string &name,
                      const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset,
                      entry_list_t &entries);
static void add_generic_entry(ifd_type_t ifd_type,
                      const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset,
                      entry_list_t &entries);

// EXIF type readers to return strings customized for bulk_extractor
static string get_exif_byte(const tiff_handle_t &tiff_handle,
                             uint32_t ifd_entry_offset, uint16_t count);
static string get_exif_ascii(const tiff_handle_t &tiff_handle,
                             uint32_t ifd_entry_offset, uint16_t count);
static string get_exif_short(const tiff_handle_t &tiff_handle,
                             uint32_t ifd_entry_offset, uint16_t count);
static string get_exif_long(const tiff_handle_t &tiff_handle,
                             uint32_t ifd_entry_offset, uint16_t count);
static string get_exif_rational(const tiff_handle_t &tiff_handle,
                             uint32_t ifd_entry_offset, uint16_t count);
static string get_exif_undefined(const tiff_handle_t &tiff_handle,
                             uint32_t ifd_entry_offset, uint16_t count);
static string get_exif_slong(const tiff_handle_t &tiff_handle,
                             uint32_t ifd_entry_offset, uint16_t count);
static string get_exif_srational(const tiff_handle_t &tiff_handle,
                             uint32_t ifd_entry_offset, uint16_t count);

exif_entry::exif_entry(uint16_t ifd_type_, const string &name_, const string &value_)
                      :ifd_type(ifd_type_), name(name_), value(value_) {
}

exif_entry::exif_entry(const exif_entry &that):ifd_type(that.ifd_type), name(that.name), value(that.value) {
}

exif_entry::~exif_entry() {
}

const string exif_entry::get_full_name() const {
    switch(ifd_type) {
    case IFD0_TIFF:
        return "ifd0.tiff." + name;
    case IFD0_EXIF:
        return "ifd0.exif." + name;
    case IFD0_GPS:
        return "ifd0.gps." + name;
    case IFD0_INTEROPERABILITY:
        return "ifd0.interoperability." + name;
    default:
        assert(0);
    }
}
 
/** 
 * parse_ifd_entries() extracts entries from an offset given its type
 */
void entry_reader::parse_ifd_entries(ifd_type_t ifd_type, const tiff_handle_t &tiff_handle,
                size_t ifd_offset, entry_list_t &entries) {
    uint32_t num_entries = ifd_reader::get_num_entries(tiff_handle, ifd_offset);
    if (num_entries == 0) {
        return;
    }

    // parse each entry
    for (uint16_t i=0; i<num_entries; i++) {
        uint32_t ifd_entry_offset = ifd_reader::get_entry_offset(ifd_offset, i);
        if (ifd_entry_offset + 12 > tiff_handle.bufsize) {
            // EOF
            break;
        } else {
            // parse the entry
            parse_entry(ifd_type, tiff_handle, ifd_entry_offset, entries);
        }
    }
}

void entry_reader::parse_entry(ifd_type_t ifd_type, const tiff_handle_t &tiff_handle,
                size_t ifd_entry_offset, entry_list_t &entries) {

    // entry tag
    const uint16_t entry_tag = tiff_reader::get_uint16(tiff_handle, ifd_entry_offset + 0);

    // add entry or parse entry's nested IFD, based on entry_tag
    const string value;
    switch (ifd_type) {
      case IFD0_TIFF:	// see table Tag Support Levels(1) for 0th IFD TIFF Tags
        uint32_t exif_ifd_offset;
        uint32_t gps_ifd_offset;
        switch (entry_tag) {
          case 0x0100: add_entry(ifd_type, "ImageWidth", tiff_handle, ifd_entry_offset, entries); break;
          case 0x0101: add_entry(ifd_type, "ImageLength", tiff_handle, ifd_entry_offset, entries); break;
          case 0x010e: add_entry(ifd_type, "ImageDescription", tiff_handle, ifd_entry_offset, entries); break;
          case 0x010f: add_entry(ifd_type, "Make", tiff_handle, ifd_entry_offset, entries); break;
          case 0x0110: add_entry(ifd_type, "Model", tiff_handle, ifd_entry_offset, entries); break;
          case 0x0131: add_entry(ifd_type, "Software", tiff_handle, ifd_entry_offset, entries); break;
          case 0x0132: add_entry(ifd_type, "DateTime", tiff_handle, ifd_entry_offset, entries); break;
          case 0x013b: add_entry(ifd_type, "Artist", tiff_handle, ifd_entry_offset, entries); break;
          case 0x8298: add_entry(ifd_type, "Copyright", tiff_handle, ifd_entry_offset, entries); break;
          case 0x8769: // EXIF tag
            exif_ifd_offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);
            entry_reader::parse_ifd_entries(IFD0_EXIF, tiff_handle, exif_ifd_offset, entries);
            break;
          case 0x8825: // GPS tag
            gps_ifd_offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);
            entry_reader::parse_ifd_entries(IFD0_GPS, tiff_handle, gps_ifd_offset, entries);
            break;
          default:
            add_generic_entry(ifd_type, tiff_handle, ifd_entry_offset, entries); break;
            break;
        }
      case IFD0_EXIF:	// see table Tag Support Levels(1) for 0th IFD TIFF Tags
        switch (entry_tag) {
          case 0x829a: add_entry(ifd_type, "ExposureTime", tiff_handle, ifd_entry_offset, entries); break;
          default:
            add_generic_entry(ifd_type, tiff_handle, ifd_entry_offset, entries); break;
        }
      case IFD0_GPS:
          add_generic_entry(ifd_type, tiff_handle, ifd_entry_offset, entries); break;
      case IFD0_INTEROPERABILITY:
          add_generic_entry(ifd_type, tiff_handle, ifd_entry_offset, entries); break;
      case IFD1_TIFF:
          add_generic_entry(ifd_type, tiff_handle, ifd_entry_offset, entries); break;
      default:
        // not a valid type
        assert(0);
    }
}

// entry value
static string get_value(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset) {
    // entry type and entry count
    const uint16_t entry_type = tiff_reader::get_uint16(tiff_handle, ifd_entry_offset + 2);
    const uint32_t entry_count = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 4);

    // value
    switch(entry_type) {
      case EXIF_BYTE: return get_exif_byte(tiff_handle, ifd_entry_offset, entry_count);
      case EXIF_ASCII: return get_exif_ascii(tiff_handle, ifd_entry_offset, entry_count);
      case EXIF_SHORT: return get_exif_short(tiff_handle, ifd_entry_offset, entry_count);
      case EXIF_LONG: return get_exif_long(tiff_handle, ifd_entry_offset, entry_count);
      case EXIF_RATIONAL: return get_exif_rational(tiff_handle, ifd_entry_offset, entry_count);
      case EXIF_UNDEFINED: return get_exif_undefined(tiff_handle, ifd_entry_offset, entry_count);
      case EXIF_SLONG: return get_exif_slong(tiff_handle, ifd_entry_offset, entry_count);
      case EXIF_SRATIONAL: return get_exif_srational(tiff_handle, ifd_entry_offset, entry_count);
      default: return "invalid entry type code " + entry_type;
    }
}

static void add_entry(ifd_type_t ifd_type, const string &name,
                      const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset,
                      entry_list_t &entries) {

    // push the name and value onto entries
    string value = get_value(tiff_handle, ifd_entry_offset);
    entries.push_back(new exif_entry(ifd_type, name, value));
}

static void add_generic_entry(ifd_type_t ifd_type,
                      const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset,
                      entry_list_t &entries) {

    // create a generic name as "entry(N)"
    const uint16_t entry_tag = tiff_reader::get_uint16(tiff_handle, ifd_entry_offset + 0);
    stringstream ss;
    ss << "entry(" << entry_tag << ")";
    string name = ss.str();

//    // add entry with generic name
//    add_entry(ifd_type, name, tiff_handle, ifd_entry_offset, entries);

    // push the name and value onto entries
    string value = get_value(tiff_handle, ifd_entry_offset);
    entries.push_back(new exif_entry(ifd_type, name, value));
}


// ************************************************************
// ifd value readers, returning string
// ************************************************************
// EXIF_BYTE uint8
static string get_exif_byte(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset, uint16_t count) {
    // get the offset to the bytes
    uint32_t offset;
    if (count <= 4) {
      offset = ifd_entry_offset + 8;
    } else {
      offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);
    }

    // get the bytes
    stringstream ss;
    int32_t working_count = (offset + count > tiff_handle.bufsize) ? tiff_handle.bufsize - offset : count;
    if (working_count <= 0) return "";
    for (int i=0; i<working_count; i++) {
      ss << tiff_handle.buf[offset + i];
//      ss << (byte)tiff_buffer[offset + i];
      if (i < working_count) {
        ss << " ";
      }
    }

    return ss.str();
}

// EXIF_ASCII byte with 7-bit ascii
static string get_exif_ascii(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset, uint16_t count) {
    // get the offset to the bytes
    uint32_t offset;
    if (count <= 4) {
      offset = ifd_entry_offset + 8;
    } else {
      offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);
    }

    // get the bytes
    int32_t working_count = (offset + count > tiff_handle.bufsize) ? tiff_handle.bufsize - offset : count;
    if (working_count <= 0) return "";
    return string(tiff_handle.buf[offset], working_count);
}

// EXIF_SHORT uint16
static string get_exif_short(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset, uint16_t count) {
    // get the offset to the bytes
    uint32_t offset;
    if (count <= 2) {
      offset = ifd_entry_offset + 8;
    } else {
      offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);
    }

    // get the shorts
    stringstream ss;
    int32_t working_count = (offset + count*2 > tiff_handle.bufsize) ? (tiff_handle.bufsize - offset)/2 : count;
    if (working_count <= 0) return "";
    for (int i=0; i<working_count; i++) {
      ss << tiff_reader::get_uint16(tiff_handle, offset + i * 2);
      if (i < working_count) {
        ss << " ";
      }
    }

    return ss.str();
}

// EXIF_LONG uint32
static string get_exif_long(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset, uint16_t count) {
    // get the offset to the bytes
    uint32_t offset;
    if (count == 1) {
      offset = ifd_entry_offset + 8;
    } else {
      offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);
    }

    // get the longs
    stringstream ss;
    int32_t working_count = (offset + count*4 > tiff_handle.bufsize) ? (tiff_handle.bufsize - offset)/4 : count;
    if (working_count <= 0) return "";
    for (int i=0; i<working_count; i++) {
      ss << tiff_reader::get_uint32(tiff_handle, offset + i * 4);
      if (i < working_count) {
        ss << " ";
      }
    }

    return ss.str();
}

// EXIF_RATIONAL uint64
static string get_exif_rational(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset, uint16_t count) {
    // get the offset to the bytes
    uint32_t offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);

    // get the rationals
    stringstream ss;
    int32_t working_count = (offset + count*8 > tiff_handle.bufsize) ? (tiff_handle.bufsize - offset)/8 : count;
    if (working_count <= 0) return "";
    for (int i=0; i<working_count; i++) {
      ss << tiff_reader::get_uint64(tiff_handle, offset + i * 8);
      if (i < working_count) {
        ss << " ";
      }
    }
    return ss.str();
}

// EXIF_UNDEFINED byte whose value depends on the field definition
static string get_exif_undefined(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset, uint16_t count) {
    return get_exif_byte(tiff_handle, ifd_entry_offset, count);
}

// EXIF_SLONG int32
static  string get_exif_slong(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset, uint16_t count) {
    // get the offset to the bytes
    uint32_t offset;
    if (count == 1) {
      offset = ifd_entry_offset + 8;
    } else {
      offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);
    }

    // get the longs
    stringstream ss;
    int32_t working_count = (offset + count*4 > tiff_handle.bufsize) ? (tiff_handle.bufsize - offset)/4 : count;
    if (working_count <= 0) return "";
    for (int i=0; i<working_count; i++) {
      ss << tiff_reader::get_int32(tiff_handle, offset + i * 4);
      if (i < working_count) {
        ss << " ";
      }
    }

    return ss.str();
}

// EXIF_SRATIONAL int64
static string get_exif_srational(const tiff_handle_t &tiff_handle, uint32_t ifd_entry_offset, uint16_t count) {
    // get the offset to the bytes
    uint32_t offset = tiff_reader::get_uint32(tiff_handle, ifd_entry_offset + 8);

    // get the rationals
    stringstream ss;
    int32_t working_count = (offset + count*8 > tiff_handle.bufsize) ? (tiff_handle.bufsize - offset)/8 : count;
    if (working_count <= 0) return "";
    for (int i=0; i<working_count; i++) {
      ss << tiff_reader::get_int64(tiff_handle, offset + i * 8);
      if (i < working_count) {
        ss << " ";
      }
    }
    return ss.str();
}

#endif
