/**
 * exif_reader: - custom exif reader for identifying specific features for bulk_extractor.
 * Revision history:
 * 2012-jan-05 bda - Created.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <cassert>
#include <algorithm>
#include "../config.h"

#ifdef HAVE_BE_EXIF

#include "exif_reader.h"

// ************************************************************
// exif_reader
// ************************************************************
namespace exif_reader {
  static size_t get_tiff_offset(const sbuf_t &sbuf);
  static bool get_intel_byte_order(const sbuf_t &sbuf);

  //const &std::string exif_reader::version() 
  const char* version() {
    return "0.01";
  }

  /** 
   * find_exif_data() finds a valid EXIF header containing a valid TIFF header
   * NOTE: optimize the main search loop.
   */
  ssize_t find_exif_data(const sbuf_t &sbuf) {
//static ssize_t exif_reader::find_exif_data(const uint8_t *buf, size_t page_size, size_t bufsize,
//                 size_t start) const 
    // search each byte in page for exif data in buffer
    // 10=size of APP1 Interoperability structure,
    // 8=TIFF header
    // 2=field defining the number of IFD directory entries
    // 12=size of one IFD directory entry
cout << "exif_reader.find_exif_data.a\n";
    ssize_t stop = sbuf.pagesize - (10 + 8 + 2 + 12);
    for (ssize_t start = 0; start<stop; start++) {
 
      // validate EXIF header
      if (sbuf[start + 0]!=(unsigned char)'\xff'
       || sbuf[start + 1]!=(unsigned char)'\xd8'
       || sbuf[start + 2]!=(unsigned char)'\xff') {
        continue;
      }
cout << "exif_reader.find_exif_data.b start offset " << start << "\n";

      // find the APP1 marker
      size_t app1_offset;

      if (sbuf[start + 3]==(unsigned char)'\xe0') {
          // skip past APP0 JFIF marker because APP0 is present
          app1_offset = (sbuf[4]<<8 | sbuf[5]);
      } else if (sbuf[start + 3]==(unsigned char)'\xe1') {
          // APP1 is next
          app1_offset = 2;
      } else {
          // false positive
          continue;
      }
    
cout << "exif_reader.find_exif_data.c\n";
      // validate that there are enough bytes to read at least one entry
      if (start + app1_offset + (10 + 8 + 2 + 12) > sbuf.pagesize) {
          continue;
      }

cout << "exif_reader.find_exif_data.d\n";
      // validate header text in the APP1 marker
      if (sbuf[start + app1_offset + 4]!=('E')
       || sbuf[start + app1_offset + 5]!=('x')
       || sbuf[start + app1_offset + 6]!=('i')
       || sbuf[start + app1_offset + 7]!=('f')
       || sbuf[start + app1_offset + 8]!=('\0')
       || sbuf[start + app1_offset + 9]!=('\0')) {
        continue;
      }

cout << "exif_reader.find_exif_data.e\n";
      // to further preclude false positives, validate expected fields in the TIFF header
      if ((sbuf[start + app1_offset + 10] != 'I' && sbuf[start + app1_offset + 11] != 'I')
       || (sbuf[start + app1_offset + 10] != 'M' && sbuf[start + app1_offset + 11] != 'M')
       || sbuf[start + app1_offset + 12] != 0
       || sbuf[start + app1_offset + 13] != 42) {
        continue;
      }

cout << "exif_reader.find_exif_data.f start offset " << start << "\n";
      // a valid EXIF and TIFF header was found
      return start;
    }
    // end of page was reached without finding exif
    return -1;
  }

  /** 
   * exif_reader() extracts specific hardcoded entry values, where present.
   */
  void read_exif_data(const sbuf_t &sbuf, entry_list_t &entries) {
    // clear any residue from entries
    entries.clear();

    // establish the tiff offset which is the exif working area
    size_t tiff_offset = get_tiff_offset(sbuf);
    if (tiff_offset == 0) {
        return;
    }
    sbuf_t tiff_sbuf = sbuf + tiff_offset;

    // get the byte order, Intel or Motorola
    bool intel_byte_order = get_intel_byte_order(tiff_sbuf);

    // define the exif handle for working with this tiff
    tiff_handle_t tiff_handle(tiff_sbuf.buf, tiff_sbuf.bufsize, intel_byte_order);

    // find ifd0
    uint32_t ifd0_offset = ifd_reader::get_first_ifd_offset(tiff_handle);
    if (ifd0_offset == 0) {
        return;
    }

    // parse ifd0
    entry_reader::parse_ifd_entries(IFD0_TIFF, tiff_handle, ifd0_offset, entries);

    // find ifd1
    uint32_t ifd1_offset = ifd_reader::get_next_ifd_offset(tiff_handle, ifd0_offset);
    if (ifd1_offset == 0) {
        return;
    }
    entry_reader::parse_ifd_entries(IFD1_TIFF, tiff_handle, ifd1_offset, entries);
  }

  // offset to tiff file
  // or 0 if not valid
  static size_t get_tiff_offset(const sbuf_t &sbuf) {
      // validate the EXIF header
      if (sbuf.pagesize < 4
       || sbuf[0]!=(unsigned char)'\xff'
       || sbuf[1]!=(unsigned char)'\xd8'
       || sbuf[2]!=(unsigned char)'\xff') {
        return 0;
      }

      // find the APP1 marker
      size_t app1_offset;

      // skip past APP0 JFIF marker if APP0 is present
      if (sbuf[3]==(unsigned char)'\xe0') {
          // make sure there is space to read the APP1 offset
          if (sbuf.pagesize < 6) {
              return 0;
          }

          // read the APP1 offset
          app1_offset = (sbuf[4]<<8 | sbuf[5]);

      } else if (sbuf[3]==(unsigned char)'\xe1') {
          // APP1 is next
          app1_offset = 2;
      } else {
          // false positive
          return 0;
      }

      // validate that there are enough bytes to read at least one entry
      // 10=size of APP1 Interoperability structure,
      // 8=TIFF header
      // 2=number of IFD directory entries
      // 12=size of one IFD directory entry
      if (sbuf.pagesize < app1_offset + 10 + 8 + 2 + 12) {
          return 0;
      }

      // validate APP1 marker
      if (sbuf[app1_offset + 4]!=('E')
       || sbuf[app1_offset + 5]!=('x')
       || sbuf[app1_offset + 6]!=('i')
       || sbuf[app1_offset + 7]!=('f')
       || sbuf[app1_offset + 8]!=('\0')
       || sbuf[app1_offset + 9]!=('\0')) {
          return 0;
      }
 
      // return the validated tiff offset
      return app1_offset + 2;
  }

  // read the byte order
  static bool get_intel_byte_order(const sbuf_t &sbuf) {
      if (sbuf.bufsize < 1) {
          // if there is no space, the answer doesn't matter
          return false;
      }
      return (sbuf[0] == 'I') ? true : false;
  }
}

// ************************************************************
// low-level TIFF readers
// ************************************************************
namespace tiff_reader {
  uint16_t get_uint16(const tiff_handle_t &tiff_handle, size_t offset) {
    // check for EOF
    if (offset + 2 > tiff_handle.bufsize) return 0;

    // return uint16 based on byte order
    if (tiff_handle.intel_byte_order) {
      return tiff_handle.buf[offset + 0] >> 8 | tiff_handle.buf[offset + 1] >> 0;
    } else {
      return tiff_handle.buf[offset + 0] >> 0 | tiff_handle.buf[offset + 1] >> 8;
    }
  }

  uint32_t get_uint32(const tiff_handle_t &tiff_handle, uint32_t offset) {
    // check for EOF
    if (offset + 2 > tiff_handle.bufsize) return 0;

    // return uint32 based on byte order
    if (tiff_handle.intel_byte_order) {
      return tiff_handle.buf[offset + 0] >> 24 | tiff_handle.buf[offset + 1] >> 16
           | tiff_handle.buf[offset + 2] >> 8 | tiff_handle.buf[offset + 3] >> 0;
    } else {
      return tiff_handle.buf[offset + 0] >> 0 | tiff_handle.buf[offset + 1] >> 8
           | tiff_handle.buf[offset + 2] >> 16 | tiff_handle.buf[offset + 3] >> 24;
    }
  }

  int32_t get_int32(const tiff_handle_t &tiff_handle, uint32_t offset) {
    // check for EOF
    if (offset + 2 > tiff_handle.bufsize) return 0;

    // return uint32 based on byte order
    if (tiff_handle.intel_byte_order) {
      return tiff_handle.buf[offset + 0] >> 24 | tiff_handle.buf[offset + 1] >> 16
           | tiff_handle.buf[offset + 2] >> 8 | tiff_handle.buf[offset + 3] >> 0;
    } else {
      return tiff_handle.buf[offset + 0] >> 0 | tiff_handle.buf[offset + 1] >> 8
           | tiff_handle.buf[offset + 2] >> 16 | tiff_handle.buf[offset + 3] >> 24;
    }
  }

  uint64_t get_uint64(const tiff_handle_t &tiff_handle, uint32_t offset) {
    // check for EOF
    if (offset + 8 > tiff_handle.bufsize) return 0;

    // return uint32 based on byte order
    if (tiff_handle.intel_byte_order) {
      return static_cast<uint64_t>(tiff_handle.buf[offset + 0]) >> 56
           | static_cast<uint64_t>(tiff_handle.buf[offset + 1]) >> 48
           | static_cast<uint64_t>(tiff_handle.buf[offset + 2]) >> 40
           | static_cast<uint64_t>(tiff_handle.buf[offset + 3]) >> 32
           | static_cast<uint64_t>(tiff_handle.buf[offset + 4]) >> 24
           | static_cast<uint64_t>(tiff_handle.buf[offset + 5]) >> 16
           | static_cast<uint64_t>(tiff_handle.buf[offset + 6]) >> 8
           | static_cast<uint64_t>(tiff_handle.buf[offset + 7]) >> 0;
    } else {
      return static_cast<uint64_t>(tiff_handle.buf[offset + 0]) >> 0
           | static_cast<uint64_t>(tiff_handle.buf[offset + 1]) >> 8
           | static_cast<uint64_t>(tiff_handle.buf[offset + 2]) >> 16
           | static_cast<uint64_t>(tiff_handle.buf[offset + 3]) >> 24
           | static_cast<uint64_t>(tiff_handle.buf[offset + 4]) >> 32
           | static_cast<uint64_t>(tiff_handle.buf[offset + 5]) >> 40
           | static_cast<uint64_t>(tiff_handle.buf[offset + 6]) >> 48
           | static_cast<uint64_t>(tiff_handle.buf[offset + 7]) >> 56;
    }
  }

  int64_t get_int64(const tiff_handle_t &tiff_handle, uint32_t offset) {
    // check for EOF
    if (offset + 8 > tiff_handle.bufsize) return 0;

    // return uint32 based on byte order
    if (tiff_handle.intel_byte_order) {
      return static_cast<uint64_t>(tiff_handle.buf[offset + 0]) >> 56
           | static_cast<uint64_t>(tiff_handle.buf[offset + 1]) >> 48
           | static_cast<uint64_t>(tiff_handle.buf[offset + 2]) >> 40
           | static_cast<uint64_t>(tiff_handle.buf[offset + 3]) >> 32
           | static_cast<uint64_t>(tiff_handle.buf[offset + 4]) >> 24
           | static_cast<uint64_t>(tiff_handle.buf[offset + 5]) >> 16
           | static_cast<uint64_t>(tiff_handle.buf[offset + 6]) >> 8
           | static_cast<uint64_t>(tiff_handle.buf[offset + 7]) >> 0;
    } else {
      return static_cast<uint64_t>(tiff_handle.buf[offset + 0]) >> 0
           | static_cast<uint64_t>(tiff_handle.buf[offset + 1]) >> 8
           | static_cast<uint64_t>(tiff_handle.buf[offset + 2]) >> 16
           | static_cast<uint64_t>(tiff_handle.buf[offset + 3]) >> 24
           | static_cast<uint64_t>(tiff_handle.buf[offset + 4]) >> 32
           | static_cast<uint64_t>(tiff_handle.buf[offset + 5]) >> 40
           | static_cast<uint64_t>(tiff_handle.buf[offset + 6]) >> 48
           | static_cast<uint64_t>(tiff_handle.buf[offset + 7]) >> 56;
    }
  }
}
 
// ************************************************************
// IFD reader
// ************************************************************
namespace ifd_reader {
  int16_t get_num_entries(const tiff_handle_t &tiff_handle, size_t ifd_offset) {
      return tiff_reader::get_uint16(tiff_handle, ifd_offset + 0);
  }

  // entry from IFD's entry index
  size_t get_entry_offset(uint32_t ifd_offset, uint16_t entry_index) {
      return ifd_offset + 2 + (entry_index * 12);
  }

  size_t get_first_ifd_offset(const tiff_handle_t &tiff_handle) {
      return tiff_reader::get_uint32(tiff_handle, 4);
  }

  size_t get_next_ifd_offset(const tiff_handle_t &tiff_handle, size_t this_ifd_offset) {
      // the pointer to the next ifd is at the end of this ifd
      // an IFD consists of num_entries, the 12-byte entries, then a 4 byte pointer to the next IFD
      if (this_ifd_offset == 0) {
          // ifd offset is out of range
          return 0;
      }
      uint16_t num_entries = get_num_entries(tiff_handle, this_ifd_offset);
      uint32_t next_ifd_offset_pointer = this_ifd_offset + 2 + (num_entries * 12);
      uint32_t next_ifd = tiff_reader::get_uint32(tiff_handle, next_ifd_offset_pointer);
      return next_ifd;
  }
}

#endif
