#include "bulk_extractor.h"

#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif

#ifndef MAXPATHLEN
#define MAXPATHLEN 65536
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif




size_t feature_recorder::context_window=16;			/* number of bytes of context */
int64_t feature_recorder::offset_add   = 0;
string feature_recorder::banner_file;


/* Truncate a string after a character */
void truncate_at(string &line,char ch)
{
    size_t pos = line.find(ch);
    if(pos!=string::npos) line.erase(pos);
}


/* 
 * Read the contents of opt_banner and put each line into "f"
 */

void feature_recorder::banner_stamp(std::ostream &o)
{
    if(banner_file.size()==0) return;	// no banner
    ifstream i(banner_file.c_str());
    if(i.is_open()){
	string line;
	getline(i,line);
	o << "#" << line << "\n";
    }
    i.close();
}


/**
 * Create a feature recorder object. Each recorder records a certain
 * kind of feature.  Features are stored in a file. The filename is
 * permutated based on the total number of threads and the current
 * thread that's recording. Each thread records to a different file,
 * and thus a different feature recorder, to avoid locking
 * problems. 
 *
 * @param name - the name of the feature being recorded.
 * @param histogram_enabled - whether or not a histogram should be created.
 */

feature_recorder::feature_recorder(string outdir_,string name_):
    flags(0),histogram_enabled(false),quoting_enabled(true),
    outdir(outdir_),name(name_),count(0),ios(),Mf(),Mr(), 
    stop_list_recorder(0),carved_set(),
    file_number(0),file_extension("."+name)
{
}

/* Don't have to erase the stop_list_recorder because it is in the set */
feature_recorder::~feature_recorder()
{
    if(ios.is_open()){
	ios.close();
    }
}


/**
 * Return the filename with a counter
 */
string feature_recorder::fname_counter(string suffix)
{
    return outdir + "/" + this->name + (suffix.size()>0 ? (string("_") + suffix) : "") + ".txt";
}


/**
 * open a feature recorder file in the specified output directory.
 */

void feature_recorder::open()
{ 
    string fname = fname_counter("");
    ios.open(fname.c_str(),ios_base::in|ios_base::out|ios_base::ate);
    if(ios.is_open()){			// opened existing stream
	ios.seekg(0L,ios_base::end);
	while(ios.is_open()){
	    /* Get current position */
	    if(int(ios.tellg())==0){		// at beginning of file; stamp and return
		ios.seekp(0L,ios_base::beg);	// be sure we are at the beginning of the file
		return;
	    }
	    ios.seekg(-1,ios_base::cur); // backup to once less than the end of the file
	    if (ios.peek()=='\n'){	     // we are finally on the \n
		ios.seekg(1L,ios_base::cur); // move the getting one forward
		ios.seekp(ios.tellg(),ios_base::beg); // put the putter at the getter location
		count = 1;			      // greater than zero
		return;
	    }
	}
    }
    // Just open the stream for output
    ios.open(fname.c_str(),ios_base::out);
    if(!ios.is_open()){
	cerr << "*** CANNOT OPEN FEATURE FILE " << fname << ":" << strerror(errno) << "\n";
	exit(1);
    }
}

void feature_recorder::close()
{
    if(ios.is_open()){
	ios.close();
    }
}

void feature_recorder::flush()
{
    cppmutex::lock lock(Mf);		// get the lock; released when object is deallocated.
    ios.flush();
}


static inline bool isoctal(char c)
{
    return c>='0' && c<='7';
}

/**
 * Unquote octal-style quoting of a string
 */
string feature_recorder::unquote_string(const string &s)
{
    size_t len = s.size();
    if(len<4) return s;			// too small for a quote

    string out;
    for(size_t i=0;i<len;i++){
	if(i+3<len && s[i]=='\\' && isoctal(s[i+1]) && isoctal(s[i+2]) && isoctal(s[i+3])){
	    uint8_t code = (s[i+1]-'0') * 64 + (s[i+2]-'0') * 8 + (s[i+3]-'0');
	    out.push_back(code);
	    i += 3;			// skip over the digits
	    continue;
	}
	out.push_back(s[i]);
    }
    return out;
}

/**
 * Get the feature which is defined as being between a \t and [\t\n]
 */

/*static*/ string feature_recorder::extract_feature(const string &line)
{
    size_t tab1 = line.find('\t');
    if(tab1==string::npos) return "";	// no feature
    size_t feature_start = tab1+1;
    size_t tab2 = line.find('\t',feature_start);
    if(tab2!=string::npos) return line.substr(feature_start,tab2-feature_start);
    return line.substr(feature_start);	// no context to remove
}

/**
 *  Create a histogram for this feature recorder and an extraction pattern.
 */
static const int max_histograms = 10;
void feature_recorder::make_histogram(string pattern,string suffix)
{
    myregex reg(pattern,REG_EXTENDED);
    string ifname = fname_counter("");	// source of features

    ifstream f(ifname.c_str());
    if(!f.is_open()){
	cerr << "Cannot open histogram input file: " << ifname << "\n";
	return;
    }
    for(int histogram_counter = 0;histogram_counter<max_histograms;histogram_counter++){
	HistogramMaker h;		/* of seen features, created in pass two */
	try {
	    string line;
	    while(getline(f,line)){
		if(line.size()==0) continue; // empty line
		if(line[0]=='#') continue;   // comment line
		truncate_at(line,'\r');	     // truncate at a \r if there is one.

		string feature = extract_feature(line);
		if(feature.find('\\')!=string::npos){
		    feature = unquote_string(feature);	// reverse \xxx encoding
		}
		/** If there is a pattern to use to prune down the feature, use it */
		if(pattern.size()){
		    string new_feature = feature;
		    if(!reg.search(feature,&new_feature,0,0)){
			continue;		// no search match; avoid this feature
		    }
		    feature = new_feature;
		}
	
		/* Remove what follows after \t if this is a context file */
		size_t tab=feature.find('\t');
		if(tab!=string::npos) feature.erase(tab); // erase from tab to end
		h.add(feature);
	    }
	    f.close();
	}
	catch (const std::exception &e) {
	    cerr << "ERROR: " << e.what() << " generating histogram " << name << "\n";
	}
	    
	/* Output what we have */
	stringstream real_suffix;

	real_suffix << suffix;
	if(histogram_counter>0) real_suffix << histogram_counter;
	string ofname = fname_counter(real_suffix.str()); // histogram name
	ofstream o;
	o.open(ofname.c_str());
	if(!o.is_open()){
	    cerr << "Cannot open histogram output file: " << ofname << "\n";
	    return;
	}
	HistogramMaker::FrequencyReportVector *fr = h.makeReport();
	if(fr->size()>0){
	    banner_stamp(o);
	    o << "# Histogram File Version: 1.1\n";
	    o << *fr;			// sends the entire histogram
	}
	delete fr;
	o.close();
	if(f.is_open()==false) return;	// input file was closed
    }
    cerr << "Looped " << max_histograms << " times on histogram; something seems wrong\n";
}

/**
 * quote a string such that everything is printable.
 * characters less than ' ' are turned into \xxx where xxx is the octal representation.
 * \\ is turned into \134
 * high-byte characters are left as-is
 */

string feature_recorder::quote_string(const string &feature)
{
    if(quoting_enabled==false) return feature;	// no quoting
    string res;
    for(string::const_iterator it = feature.begin(); it!=feature.end(); it++){
	if((*it >= '\000' && *it<' ') || *it=='\\'){
	    char buf[6];
	    snprintf(buf,sizeof(buf),"\\%03o",*it);
	    res.append(buf);
	} else {
	    res.push_back(*it);
	}
    }
    return res;
}

/* static */ const string feature_recorder::feature_file_header("# Feature File Version: 1.1\n");

/* Write to the file.
 * This is the only place where writing happens.
 */
void feature_recorder::write(const std::string &str)
{
    cppmutex::lock lock(Mf);
    if(ios.is_open()){
	if(count==0){
	    banner_stamp(ios);
	    ios << feature_file_header;
	}

	ios << str << '\n';
	if(ios.fail()){
	    cerr << "DISK FULL\n";
	    ios.close();
	}
	count++;
    }
}

void feature_recorder::printf(const char *fmt, ...)
{
    const int maxsize = 65536;
    char *p;
    va_list ap;
    if ((p = (char *)malloc(maxsize))==0) return;
    va_start(ap,fmt);
    vsnprintf(p,maxsize,fmt,ap);
    va_end(ap);
    this->write(p);
    free(p);
}


/**
 * the main entry point of writing a feature and its context to the feature file.
 * processes the stop list
 */

void feature_recorder::write(const pos0_t &pos0,string feature,string context)
{
    if(flags & FLAG_DISABLED) return;		// disabled

    /* non-printable junk. This is why references can't be passed in */
    feature = quote_string(feature);
    if((flags & FLAG_NO_CONTEXT)==0){
	context = quote_string(context);
    }

    if(feature.size()==0){
	cerr << "zero length feature at " << pos0 << "\n";
	return;
    }

    /* First check to see if the feature is on the stop list.
     * Only do this if we have a stop_list_recorder (the stop list recorder itself
     * does not have a stop list recorder. If it did we would infinitely recurse.
     */

    if(((flags & FLAG_NO_STOPLIST)==0) && stop_list_recorder){		
	if(stop_list.check_feature_context(feature,context)){
	    stop_list_recorder->write(pos0,feature,context);
	    return;
	}
    }

    /* The alert list is a special features that are called out.
     * If we have one of those, write it to the redlist.
     */
    if(((flags & FLAG_NO_ALERTLIST)==0) && alert_list.check_feature_context(feature,context)){
	string alert_fn = outdir + "/ALERTS_found.txt";

	cppmutex::lock lock(Mr);		// notce we are locking the redlist
	ofstream rf(alert_fn.c_str(),ios_base::app);
	if(rf.is_open()){
	    rf << pos0.shift(feature_recorder::offset_add).str() << '\t' << feature << '\t' << "\n";
	}
    }

    /* Finally write out the feature and the context */
    stringstream ss;
    ss << pos0.shift(feature_recorder::offset_add).str() << '\t' << feature;
    if(((flags & FLAG_NO_CONTEXT)==0) && (context.size()>0)) ss << '\t' << context;
    this->write(ss.str());
}

/**
 * Given a buffer, an offset into that buffer of the feature, and the length
 * of the feature, make the context and write it out. This is mostly used
 * for writing from within the lexical analyzers.
 */

void feature_recorder::write_buf(const sbuf_t &sbuf,size_t pos,size_t len)
{
    /* If we are in the margin, ignore; it will be processed again */
    if(pos >= sbuf.pagesize && pos < sbuf.bufsize){
	return;
    }

    if(pos >= sbuf.bufsize){    /* Sanity checks */
	cerr << "*** write_buf: WRITE OUTSIDE BUFFER.  pos="
	     << pos
	     << " sbuf=" << sbuf << "\n";
	return;
    }

    /* Asked to write beyond bufsize; bring it in */
    if(pos+len > sbuf.bufsize){
	len = sbuf.bufsize - pos;
    }

    string feature = string((const char *)sbuf.buf+pos,len);
    string context;

    if((flags & FLAG_NO_CONTEXT)==0){
	/* Context write; create a clean context */
	size_t p0 = context_window < pos ? pos-context_window : 0;
	size_t p1 = pos+len+context_window;
	
	if(p1>sbuf.bufsize) p1 = sbuf.bufsize;
	
	assert(p0<=p1);
	context = sbuf.substr(p0,p1-p0);
    }
    this->write(sbuf.pos0+pos,feature,context);
}


/**
 * Carving support.
 */
void feature_recorder::carve(const sbuf_t &sbuf,size_t pos,size_t len)
{
    /* If we are in the margin, ignore; it will be processed again */
    if(pos >= sbuf.pagesize && pos < sbuf.bufsize){
	return;
    }

    if(pos >= sbuf.bufsize){    /* Sanity checks */
	cerr << "*** carve: WRITE OUTSIDE BUFFER.  pos=" << pos
	     << " sbuf=" << sbuf << "\n";
	return;
    }

    
    /* If the directory doesn't exist, make it. If two threads try to make the directory,
     * that's okay, because the second one will fail.
     */

    string dirname = outdir + "/" + name;
    if (access(dirname.c_str(),R_OK)!=0) be_mkdir(dirname);

    /* Get the MD5 of the object to be carved */
    md5_t carved_md5 = sbuf.md5(pos,len);
    char fname[MAXPATHLEN+1];
    
    {
	/* New code: grab the lock. It is automatically freed on return. */
	cppmutex::lock lock(Mf);
	bool previously_carved = (carved_set.find(carved_md5)!=carved_set.end());

	if(previously_carved) return;

	/* Insert it into the set */
	carved_set.insert(carved_md5);

	/* Generate the file name */
	int myfilenumber = file_number++;
	snprintf(fname,sizeof(fname),"%s/%06d%s",dirname.c_str(),myfilenumber,file_extension.c_str());
	/* mutex must be freed before call to write, since write() will lock it again */
    }
    
    /* Record what was found in the feature file.
     */
    char buf[MAXPATHLEN*2+1];
    snprintf(buf,sizeof(buf),
	     "<fileobject><filename>%s</filename><filesize>%d</filesize>"
	     "<hashdigest type='MD5'>%s</hashdigest></fileobject>",
	     fname,(int)len,carved_md5.hexdigest().c_str());
    this->write(sbuf.pos0+len,fname,buf);
    
    /* Now carve to a file */
    errno=0;
    int fd = ::open(fname,O_CREAT|O_BINARY|O_RDWR,0666);
    if(fd<0){
	cerr << "*** carve: Cannot create " << buf << ": " << strerror(errno) << "\n";
	return;
    }
    ssize_t ret = sbuf.write(fd,pos,len);
    if(ret!=(ssize_t)len){
	cerr << "*** carve: Cannot write(" << fd << "," << pos
	     << "," << len << ")=" << ret
	     << ": "<< strerror(errno) << "\n";
    }
    ::close(fd);
}

