/**
 * feature_recorder.h:
 *
 * System for recording features from the scanners into the feature files.
 *
 * This module defines three classes:
 * class ifstream_line - an ifstream that can read a line at a time and remember offsets.
 * class feature_recorder - the individual recorders.
 * class feature_recorder_set - the set of recorders.
 * 
 * There is one feature_recorder per feature file. It is used both to record
 * the features and to perform the histogram calculation.
 * (That should probably be moved to a different class.) It also also previously
 * had the ability to do a merge sort, but we took that out because it was
 * not necessary.
 *
 * The feature recorders can also check the global alert_list to see
 * if the feature should be written to the alert file. It's opened on
 * demand and immediately flushed and closed.  A special mutex is used
 * to protect it.
 *
 * Finally, the feature recorder supports the global stop_list, which
 * is a list of features that are not written to the main file but are
 * written to a stop list.  That is implemented with a second
 * feature_recorder.
 *
 * There is one feature_recorder_set per process.
 * The file assumes that bulk_extractor.h is being included.
 */
 
#ifndef FEATURE_RECORDER_H
#define FEATURE_RECORDER_H

using namespace std;
#include <string>
#include <pthread.h>
#include <stdarg.h>

#include "md5.h"
#include "regex_list.h"
#include "histogram.h"

void truncate_at(string &line,char ch);
class feature_recorder {
private:
    uint32_t flags;
    bool histogram_enabled;		/* do we automatically histogram? */
    bool quoting_enabled;		// by default, quote invalid ASCII characters in output
    /*** neither copying nor assignment is implemented ***
     *** We do this by making them private constructors that throw exceptions. ***/
    class not_impl: public exception {
	virtual const char *what() const throw() {
	    return "copying feature_recorder objects is not implemented.";
	}
    };
    feature_recorder(const feature_recorder &fr):
	flags(0),histogram_enabled(false),quoting_enabled(false),
	outdir(),name(),count(0),ios(),Mf(),Mr(),
	stop_list_recorder(0),carved_set(),file_number(0),file_extension(){ throw new not_impl(); }
    const feature_recorder &operator=(const feature_recorder &fr){ throw new not_impl(); }
    /****************************************************************/


public:
    static const int FLAG_DISABLED=0x01;	// don't do anything
    static const int FLAG_NO_CONTEXT=0x02;	// do not write context
    static const int FLAG_NO_STOPLIST=0x04;	// do not honor stoplist
    static const int FLAG_NO_ALERTLIST=0x04;	// do not honor stoplist
    static const string feature_file_header;
    typedef string offset_t;
    static size_t context_window;	// global option
    static int64_t offset_add;		// added to every reported offset
    static string banner_file;
    static void   banner_stamp(std::ostream &os);
    static string extract_feature(const string &line);
    static bool histogram_add_file(string fname,string pattern,
				   set<offset_t>&h_offsets,HistogramMaker &h);

    feature_recorder(string outdir,string name);
    virtual ~feature_recorder();

    void set_flag(uint32_t flags_){flags|=flags_;}

    string outdir;			// where output goes
    string name;			/* name of this feature recorder */
    int64_t count;			/* number of records written */
    std::fstream ios;			/* where features are written */
    cppmutex Mf;			/* protects the file */
    cppmutex Mr;			/* protects the redlist */

    /* where stopped items (on stop_list or context_stop_list) get recorded: */
    class feature_recorder *stop_list_recorder; // where stopped features get written
    string fname_counter(string suffix);
    string quote_string(const string &feature); // turns unprintable characters to octal escape
    static string unquote_string(const string &feature); // turns octal escape back to binary characters

    void open();
    void close();			
    void flush();
    void make_histogram(string pattern,string basename);
    
    /* Methods to write.
     * write() is the basic write - you say where, and it does it.
     * write_buf() writes from a position within the buffer, with context.
     *             It won't write a feature that starts in the margin.
     * pos0 gives the location and prefix for the beginning of the buffer
     */ 

    /**
     * write actually does the writing with locks.
     * All others others call write and therefore do not need locks.
     */
    void write(const std::string &str);

    /**
     * support for carving.
     * Carving writes the filename to the feature file; the context is the file's MD5 
     */
    set<md5_t>	carved_set;		/* set of MD5 hash codes of objects we've carved */
    int64_t	file_number;		/* starts at 0 */
    string	file_extension;		/* includes "." */
    void carve(const sbuf_t &sbuf,size_t pos,size_t len);

    // virtual functions may be called by plug-ins
    // printf() prints to the feature file.
    virtual void printf(const char *fmt_,...) __attribute__((format(printf, 2, 3)));
    // 
    // write a feature and its context; the feature may be in the context, but doesn't need to be.
    virtual void write(const pos0_t &pos0,string feature,string context);  

    // write a feature located at a given place within an sbuf.
    // Context is written automatically
    virtual void write_buf(const sbuf_t &sbuf,size_t pos,size_t len); /* writes with context */

    // Enable or disable quoting; turning it off improves performance, but makes
    // it possible to put invalid characters in the file.
    void set_quoting_enabled(bool flag){quoting_enabled=flag;}
};

#endif
