#ifndef HISTOGRAM_H
#define HISTOGRAM_H

/* C++ Histogram class.
 * The basis of a string-based correlator and many other features.
 * Uses C++ STL for sorting and string handling.
 * 
 * Summer 2011: Now is UTF-8/UTF-16 aware. All strings are stored as UTF-8.
 * Detects UTF-16 in an add and automatically converts to UTF-8.
 * Keeps track of UTF-16 count separately from UTF-8 count.
 *
 * Oct 2011: Apparently you are not supposed to subclass the STL container classes. 
 */
		

#include <vector>
#include <map>

class  HistogramMaker  {
public:
    /** The ReportElement is used for creating the report of histogram frequencies.
     * It can be thought of as the histogram bin.
     */
    class histogramTally {
    public:
	uint32_t count;		// total strings seen
	uint32_t count16;	// total utf16 strings seen
	histogramTally():count(0),count16(0){};
	virtual ~histogramTally(){};
    };

    class ReportElement {
    public:
	ReportElement(std::string aValue,histogramTally aTally):value(aValue),tally(aTally){ }
	std::string	value;
	histogramTally  tally;
	static bool compare(const ReportElement &e1,const ReportElement &e2) {
	    if (e1.tally.count > e2.tally.count) return true;
	    if (e1.tally.count < e2.tally.count) return false;
	    return e1.value < e2.value;
	}
	virtual ~ReportElement(){};
    };

private:
    /** A HistogramMap holds the histogram while it is being computed.
     */
    typedef std::map<std::string,histogramTally> HistogramMap;
    HistogramMap h;			// holds the histogram
public:

    /**
     * Determine if a string probably has utf16.
     */
    static bool looks_like_utf16(const std::string &str,bool &little_endian); 

    HistogramMaker():h(){}
    void clear(){h.clear();}
    void add(const std::string &key);	// adds an element to the histogram

    /** makeReport() makes a report and returns a
     * FrequencyReportVector.
     */
    /** A FrequencyReportVector is a vector of report elements when the report is generatedn.
     */
    typedef std::vector<ReportElement> FrequencyReportVector;
    FrequencyReportVector *makeReport() const;	// return a report with all of them
    FrequencyReportVector *makeReport(int topN) const; // returns just the topN
    virtual ~HistogramMaker(){}
};

std::ostream & operator <<(std::ostream &os,const HistogramMaker::FrequencyReportVector &rep);

#endif
