/*
 * myregex.h:
 * 
 * simple cover for regular expression class.
 * The class allocates and frees the strings 
 */

#ifndef MYREGEX_H
#define MYREGEX_H

#include <string>
#include <iostream>
#include <assert.h>

extern "C" {
#include <regex.h>
#include <string.h>
}


using namespace std;


class myregex {
 public:
    /** Bargain-basement detector of things that might be regular expressions. */
    static bool is_regex(const string &str){
	for(string::const_iterator it = str.begin();it!=str.end();it++){
	    switch(*it){
	    case '?':
	    case '*':
	    case '.':
	    case '+':
	    case '[':
	    case '(':
		return true;
	    }
	}
	return false;
    }

    static const int MYREG_GLOB=0x1000000; /* treat as a shell glob */
    string  pat;				   /* our pattern */
    int     flags;
    regex_t *ncv;			/* compiled regex */
    myregex(const myregex &that):pat(that.pat),flags(that.flags),ncv(0){
	compile();
    }

    myregex & operator=(const myregex&that){
	if(ncv){
	    regfree(ncv);
	    free(ncv);
	    ncv = 0;
	}
	pat=that.pat;
	flags=that.flags;
	compile();
	return *this;
    }
    myregex(std::string pat_,int flags_):pat(pat_),flags(flags_),ncv(0){
	/* If we are globbing, do the glob */
	if(flags & MYREG_GLOB){
	    this->pat = "";
	    for(string::iterator it = pat_.begin();it!=pat_.end();it++){
		switch(*it){
		case '?':pat+=".";break;
		case '*':pat+=".*";break;
		default: this->pat += *it;
		}
	    }
	    flags = flags & ~MYREG_GLOB; /* remove flag */
	}
	compile();
    }
    void compile(){			// compile the regex
	if(pat.size()==0) return;
	ncv = (regex_t *)calloc(sizeof(*ncv),1);
	if(regcomp(ncv,pat.c_str(),flags | REG_EXTENDED)){
	    std::cerr << "regular expression compile error '" << pat << "'\n";
	    exit(1);
	}
    }
    ~myregex(){
	if(ncv){
	    regfree(ncv);
	    free(ncv);
	    ncv = 0;
	}
    }
    /**
     * perform a search for a single hit. If there is a group and something is found,
     * set *found to be what was found, *offset to be the starting offset, and *len to be
     * the length. Note that this only handles a single group.
     */
     int search(const std::string &line,std::string *found,size_t *offset,size_t *len) const{
	static const int REGMAX=2;
	regmatch_t pmatch[REGMAX];
	if(!ncv) return 0;
	memset(pmatch,0,sizeof(pmatch));
	int r = regexec(ncv,line.c_str(),REGMAX,pmatch,0);
	if(r==REG_NOMATCH) return 0;
	if(r!=0) return 0;		/* some kind of failure */
	/* Make copies of the first group */
	if(pmatch[1].rm_so != pmatch[1].rm_eo){
	    if(found)  *found = line.substr(pmatch[1].rm_so,pmatch[1].rm_eo-pmatch[1].rm_so);
	    if(offset) *offset = pmatch[1].rm_so;
	    if(len)    *len = pmatch[1].rm_eo-pmatch[1].rm_so;
	}
	return 1;			/* success */
    }
    /** Perform a search with an array of strings. Return 0 if success, return code if fail.*/
    int search(const std::string &line,string *matches,int REGMAX) const {
	regmatch_t *pmatch = (regmatch_t *)calloc(sizeof(regmatch_t),REGMAX+1);
	if(!ncv) return 0;
	int r = regexec(ncv,line.c_str(),REGMAX+1,pmatch,0);
	if(r==0){
	    for(int i=0;i<REGMAX;i++){
		size_t start = pmatch[i+1].rm_so;
		size_t len   = pmatch[i+1].rm_eo-pmatch[i+1].rm_so;
		matches[i] = line.substr(start,len);
	    }
	}
	free(pmatch);
	return r;
    }
};
typedef vector<myregex *> regex_vector;

#endif
