#include "bulk_extractor.h"

/****************************************************************
 *** SBUF_T
 ****************************************************************/

/**
 * Destructer is here because it needs to know if we have mmap or not.
 * By design, sbuf.h can be read without reading config.h.
 */
void sbuf_t::release()
{
#ifdef HAVE_MMAP
    if(should_unmap && buf){
	munmap((void *)buf,bufsize);
	should_unmap = false;
	buf = 0;
    }
#endif
    if(should_close && fd>0){
	close(fd);
	should_close = false;
	fd=0;
    }
    if(should_free && buf){
	free((void *)buf);
	should_free = false;
	buf = 0;
    }
    page_number = 0;
    bufsize = 0;
    pagesize = 0;
}

/**
 *  Map a file; falls back to read if mmap is not available
 */
sbuf_t *sbuf_t::map_file(string fname)
{
    int fd = open(fname.c_str(),O_RDONLY,0);
    if(fd<0) return 0;		/* cannot open file */

    struct stat st;
    if(fstat(fd,&st)){
	close(fd);
	return 0; /* cannot stat */
    }

    sbuf_t *sbuf = new sbuf_t();
    sbuf->fd = fd;
    sbuf->bufsize = st.st_size;
    sbuf->pagesize = st.st_size;

#ifdef HAVE_MMAP
    sbuf->buf = (uint8_t *)mmap(0,st.st_size,PROT_READ,MAP_FILE|MAP_SHARED,fd,0);
    sbuf->should_free  = false;
    sbuf->should_unmap = true;
    sbuf->should_close = true;
#else
    sbuf->buf = (uint8_t *)malloc(st.st_size);
    if(sbuf->buf==0){		/* malloc failed */
	delete sbuf;
	return 0;
    }
    if((size_t)read(sbuf->fd,(void *)sbuf->buf,sbuf->bufsize)!=sbuf->bufsize){
	free((void *)sbuf->buf);		/* read failed */
	delete sbuf;
	return 0;
    }
    sbuf->should_free = true;
    sbuf->should_unmap = false;
    sbuf->should_close = false;
#endif
    sbuf->pos0.path   = fname;	// set the filename
    sbuf->pos0.offset = 0;		// zero offset
    return sbuf;
}

/**
 * rawdump the sbuf to an ostream.
 */
void sbuf_t::raw_dump(std::ostream &os,uint64_t start,uint64_t len) const
{
    for(uint64_t i=start;i<start+len  && i<bufsize;i++){
	os << buf[i];
    }
}

/**
 * rawdump the sbuf to a file descriptor
 */
void sbuf_t::raw_dump(int fd2,uint64_t start,uint64_t len) const
{
    if(len>bufsize-start) len=bufsize-start; // maximum left
    uint64_t written = ::write(fd2,buf+start,len);
    if(written!=len){
      cerr << "write: cannot write sbuf.\n";
    }
}

static string hexch(unsigned char ch)
{
    char buf[4];
    snprintf(buf,sizeof(buf),"%02x",ch);
    return string(buf);
}

/**
 * hexdump the sbuf.
 */
void sbuf_t::hex_dump(std::ostream &os,uint64_t start,uint64_t len) const
{
    const size_t bytes_per_line = 32;
    for(uint64_t i=start;i<start+len && i<bufsize;i+=bytes_per_line){
	size_t spaces=0;
	for(size_t j=0;j<bytes_per_line && i+j<bufsize && i+j<start+len;j++){
	    unsigned char ch = (*this)[i+j];
	    os << hexch(ch) << " ";
	    spaces += 3;
	    if(j==bytes_per_line/2){
		os << ' ';
		spaces += 1;
	    }
	}
	for(;spaces<bytes_per_line*3+3;spaces++){
	    os << ' ';
	}
	for(size_t j=0;j<bytes_per_line && i+j<bufsize && i+j<start+len;j++){
	    unsigned char ch = (*this)[i+j];
	    if(ch>=' ' && ch<='~') os << ch;
	    else os << '.';
	}
	os << "\n";
    }
}

/* Write to a file descriptor */
ssize_t sbuf_t::write(int fd_,size_t loc,size_t len) const
{
    if(loc>=bufsize) return 0;		// cannot write
    if(loc+len>bufsize) len=bufsize-loc; // clip at the end
    return ::write(fd_,buf+loc,len);
}

/* Return a substring */
string sbuf_t::substr(size_t loc,size_t len) const
{
    if(loc>=bufsize) return string("");		// cannot write
    if(loc+len>bufsize) len=bufsize-loc; // clip at the end
    return string((const char *)buf+loc,len);
}

/* Return the md5 of a substring */
md5_t sbuf_t::md5(size_t loc,size_t len) const
{
    if(loc>=bufsize) return md5_generator::hash_buf((const unsigned char *)"",0);
    if(loc+len>bufsize) len=bufsize-loc; // clip at the end
    return md5_generator::hash_buf(buf+loc,len);
}

void sbuf_t::hex_dump(std::ostream &os) const 
{
    hex_dump(os,0,bufsize);
}

/**
 * Convert a binary blob to a hex representation
 */

#ifndef NSRL_HEXBUF_UPPERCASE
#define NSRL_HEXBUF_UPPERCASE 0x01
#define NSRL_HEXBUF_SPACE2    0x02
#define NSRL_HEXBUF_SPACE4    0x04
#endif


static int *hexcharvals = 0;
static const char *hexbuf(char *dst,int dst_len,const unsigned char *bin,int bytes,int flag)
{
    int charcount = 0;
    const char *start = dst;		// remember where the start of the string is
    const char *fmt = (flag & NSRL_HEXBUF_UPPERCASE) ? "%02X" : "%02x";

    if(hexcharvals==0){
	/* Need to initialize this */
	int i;
	hexcharvals = (int *)calloc(sizeof(int),256);
	for(i=0;i<10;i++){
	    hexcharvals['0'+i] = i;
	}
	for(i=10;i<16;i++){
	    hexcharvals['A'+i-10] = i;
	    hexcharvals['a'+i-10] = i;
	}
    }

    *dst = 0;				// begin with null termination
    while(bytes>0 && dst_len > 3){
	sprintf(dst,fmt,*bin); // convert the next byte
	dst += 2;
	bin += 1;
	dst_len -= 2;
	bytes--;
	charcount++;			// how many characters
	
	if((flag & NSRL_HEXBUF_SPACE2) ||
	   ((flag & NSRL_HEXBUF_SPACE4) && charcount%2==0))
	    *dst++ = ' ';
	    *dst   = '\000';
	    dst_len -= 1;
    }
    return start;			// return the start
}



std::ostream & operator <<(std::ostream &os,const class sbuf_t &sbuf){
    char hex[17];
    hexbuf(hex,sizeof(hex),sbuf.buf,8,0);
    os << "sbuf["   << sbuf.page_number
       << ":" << sbuf.pos0 << "," << "buf[0..8]=" << hex
       << ", bufsize=" << sbuf.bufsize << ", pagesize=" << sbuf.pagesize << "]";
    return os;
}


