#ifndef SBUF_H
#define SBUF_H

/* required per C++ standard */
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif

#include "md5.h"

using namespace std;

#include <stdlib.h>
#include <inttypes.h>
#include <string>
#include <sstream>

/****************************************************************
 *** pos0_t
 ****************************************************************/

/**
 * The pos0_t structure is used to record the forensic path of the
 * first byte of an sbuf. The forensic path can include strings associated
 * with decompressors and ordinals associated with offsets.
 * 
 * e.g., 1000-GZIP-300-BASE64-30 means go 1000 bytes into the stream,
 *       unzip, go 300 bytes into the decompressed stream, un-BASE64, and
 *       go 30 bytes into that.
 * 
 * pos0_t uses a string to hold the base path and the offset into that path
 * in a 64-bit number.  
 */

class pos0_t {
 public:
    static inline int64_t stoi64(std::string str){
	std::istringstream ss(str);
	int64_t val;
	ss >> val;
	return val;
    }

    string   path;			/* forensic path of decoders*/
    uint64_t offset;			/* location of buf[0] */
    
    pos0_t():path(""),offset(0){}
    pos0_t(string s):path(s),offset(0){}
    pos0_t(const pos0_t &obj):path(obj.path),offset(obj.offset){ }
    string str() const {
	stringstream ss;
	if(path.size()>0){
	    ss << path << "-";
	}
	ss << offset;
	return ss.str();
    }
    string firstPart() const {
	size_t p = path.find('-');
	if(p==string::npos) return string("");
	return path.substr(0,p);
    }
    string lastAddedPart() const {
	size_t p = path.rfind('-');
	if(p==string::npos) return string("");
	return path.substr(p+1);
    }
    /**
     * Return a new position that's been shifted by an offset
     */
    pos0_t shift(int64_t s) const {
	if(s==0) return *this;
	pos0_t ret;
	size_t p = path.find('-');
	if(p==string::npos){		// no path
	    ret.path="";
	    ret.offset = offset + s;
	    return ret;
	}
	/* Figure out the value of the shift */
	int64_t baseOffset = stoi64(path.substr(0,p-1));
	stringstream ss;
	ss << (baseOffset+s) << path.substr(p);
	ret.path = ss.str();
	ret.offset = offset;
	return ret;
    }
};

/* iostream support for the pos0_t */
inline std::ostream & operator <<(std::ostream &os,const class pos0_t &pos0) {
    os << "(" << pos0.path << "|" << pos0.offset << ")";
    return os;
}


/* Append a string (subdir).
 * The current offset is a prefix to the subdir.
 */
inline class pos0_t operator +(pos0_t pos0,const string &subdir) {
    stringstream ss;
    ss << pos0.offset;
    pos0.path    += (pos0.path.size()>0 ? "-" : "") + ss.str() + "-" + subdir;
    pos0.offset  = 0;
    return pos0;
};

/* Adding an offset */
inline class pos0_t operator +(pos0_t pos0,int64_t delta) {
    pos0.offset += delta;		
    return pos0;
};

/* Comparision operations */
inline bool operator <(const class pos0_t &pos0,const class pos0_t & pos1)  {
    if(pos0.path.size()==0 && pos1.path.size()==0) return pos0.offset < pos1.offset;
    if(pos0.path == pos1.path) return pos0.offset < pos1.offset;
    return pos0.path < pos1.path;
};

inline bool operator >(const class pos0_t & pos0,const class pos0_t &pos1)  {
    if(pos0.path.size()==0 && pos1.path.size()==0) return pos0.offset > pos1.offset;
    if(pos0.path == pos1.path) return pos0.offset > pos1.offset;
    return pos0.path > pos1.path;
};

inline bool operator ==(const class pos0_t & pos0,const class pos0_t &pos1) {
    return pos0.path==pos1.path && pos0.offset==pos1.offset;
};

/****************************************************************
 *** sbuf_t
 *** This structure actually holds the data.
 *** We use a pos0_t to maintain the address of the first byte.
 ***
 *** There are lots of ways for allocating an sbuf_t:
 *** - map from a file.
 *** - set from a block of memory.
 *** - a subset of an existing sbuf_t (sbuf+10 gives you 10 bytes in, and therefore 10 bytes shorter)
 ***
 *** The subf_t class remembers how the sbuf_t was allocated and
 *** automatically frees whatever resources are needed when it is freed.
 ***
 *** DANGER: You must delete sbuf_t structures First-In, Last-out, otherwise
 *** bad things can happen. (For example, if you make a subset sbuf_t from a mapped file and unmap
 *** the file, the subset will now point to unallocated memory.)
 ***
 ****************************************************************/

/**
 * this class describes the search buffer.
 */
class sbuf_t {
private:
    /* The private structures keep track of memory management */
    int    fd;				/* file this came from if mmapped file */
    bool   should_unmap;		/* munmap buffer when done */
    bool   should_free;			/* should buf be freed when this sbuf is deleted? */
    bool   should_close;		/* close(fd) when done. */
    static size_t min(size_t a,size_t b){
	return a<b ? a : b;
    }

 public:
    int     page_number;		/* used for debugging */
    pos0_t  pos0;			/* the path of buf[0] */
    uint8_t *buf;			/* start of the buffer */
    size_t  bufsize;			/* size of the buffer */
    size_t  pagesize;			/* page data; the rest is the 'margin'. pagesize <= bufsize */
    
    sbuf_t(const sbuf_t &that):fd(0),should_unmap(false),should_free(false),should_close(false),
			       page_number(that.page_number),pos0(that.pos0),buf(0),
			       bufsize(that.bufsize),pagesize(that.pagesize){
	/* Make a copy of the data */
	buf = (uint8_t *)malloc(bufsize);
	memcpy(buf,that.buf,bufsize);
    }

    sbuf_t &operator=(const sbuf_t &that) {
	release();
	/* Make a copy of the data */
	pos0 = that.pos0;
	buf = (uint8_t *)malloc(bufsize);
	bufsize = that.bufsize;
	pagesize = that.pagesize;
	memcpy(buf,that.buf,bufsize);
	return *this;
    }

    /* Allocators */
    /* Empty allocator */
    sbuf_t():fd(0),should_unmap(false),should_free(false),should_close(0),
	     page_number(0),pos0(),buf(0),bufsize(0),pagesize(0){
    }
    /* Allocate from an existing buffer */
    sbuf_t(const pos0_t &pos0_,uint8_t *buf_,size_t bufsize_,size_t pagesize_):
	fd(0), should_unmap(false), should_free(false), should_close(false),
	page_number(0),pos0(pos0_),buf(buf_),bufsize(bufsize_),
	pagesize(min(pagesize_,bufsize_ )) {
    };
    /* Allocate from an existing buffer */
    sbuf_t(const pos0_t &pos0_,uint8_t *buf_,size_t bufsize_,size_t pagesize_,bool should_free_):
	fd(0), should_unmap(false), should_free(should_free_), should_close(false),
	page_number(0),pos0(pos0_),buf(buf_),bufsize(bufsize_),
	pagesize(min(pagesize_,bufsize_)){
    };
    virtual ~sbuf_t(){
	release();
    }
    void release();			// release allocated storage

    static sbuf_t *map_file(string fname); // map a file and return a sbuf if success, 0 if failure.
    size_t size() const {return bufsize;}

    /**
     * asString - returns the sbuf as a string
     */

    string asString() const {return string((const char *)buf,bufsize);}

    /**
     * the get functions safely return an integer value for the offset of i,
     * in Intel byte order.
     */
    uint8_t get8i(size_t i) const {
	return (*this)[i];
    }
    uint16_t get16i(size_t i) const {
	return ((*this)[i]) | ((*this)[i+1]<<8);
    }
    uint32_t get32i(size_t i) const {
	return ((*this)[i]) | ((*this)[i+1]<<8)
	    | ((*this)[i+2]<<16) | ((*this)[i+3] << 24);
    }

    /**
     * The [] operator safely returns what's at index [i] or 0
     */
    uint8_t operator [](size_t i) const {
	return (i>=0 && i<bufsize) ? buf[i] : 0;
    }

    /**
     * the + operator returns a new sbuf that is i bytes in and, therefore, i bytes smaller.
     * Not we assume that pagesize is always smaller than or equal to bufsize.
     */
    sbuf_t operator +(size_t i) const {
	assert(bufsize>=pagesize);
	if(i>pagesize) i = pagesize;
	return sbuf_t(pos0+i,buf+i,bufsize-i,pagesize-i);
    }
    ssize_t find(uint8_t ch,size_t start) const {
	for(;start<pagesize;start++){
	    if(buf[start]==ch) return start;
	}
	return -1;
    }
    ssize_t find(const char *str,size_t start) const {
	for(;start<pagesize;start++){
	    bool found = true;
	    for(size_t i=0;str[i] && found;i++){
		if(start+i>=pagesize) return -1; // ran off the end
		found = (buf[start+i]==str[i]);
	    }
	    if(found) return start;
	}
	return -1;
    }
    void raw_dump(std::ostream &os,uint64_t start,uint64_t len) const;
    void raw_dump(int fd,uint64_t start,uint64_t len) const; // writes to a raw file descriptor
    void hex_dump(std::ostream &os,uint64_t start,uint64_t len) const;
    void hex_dump(std::ostream &os) const; /* dump all */
    string substr(size_t offset,size_t len) const; /* make a substring */
    ssize_t  write(int fd,size_t offset,size_t len) const; /* write to a file descriptor  */
    md5_t md5(size_t offset,size_t len) const;	   /* compute the MD5 of a subset */
    
};
std::ostream & operator <<(std::ostream &os,const class sbuf_t &sbuf);


#endif
