/**
 * scan_be_exif: - custom exif scanner for identifying specific features for bulk_extractor.
 *
 * We do not require EXIF data to begin on a 512-byte offset boundary.
 *
 * Revision history:
 * 2011-dec-12 bda - Ported from file scan_exif.cpp.
 */

#include "bulk_extractor.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <cassert>
#include <algorithm>

#ifdef HAVE_BE_EXIF

#include "md5.h"
#include "exif_reader.h"

/****************************************************************/
/* C++ string splitting code from http://stackoverflow.com/questions/236129/how-to-split-a-string-in-c */

using namespace std;

inline size_t min(size_t a,size_t b){
    return a<b ? a : b;
}

/*
inline size_t max(size_t a,size_t b)
{
    return a>b ? a : b;
}
*/

inline string format_exif_data(const entry_list_t &tags);
inline string format_gps_data(const entry_list_t &tags);

/** Return the hex md5 of the first 4k of a buffer */
static string md5hex_4k(const sbuf_t &sbuf)
{
    return md5_generator::hash_buf(sbuf.buf,min(sbuf.pagesize,4096)).hexdigest();
}

extern "C"
void scan_be_exif(const class scanner_params &sp,const recursion_control_block &rcb)
{
cout << "scan_be_exif.a\n";
    assert(sp.version==1);
    if(sp.phase==0){
	sp.info->name  = "be_exif";
	sp.info->feature_names.insert("be_exif");
	sp.info->feature_names.insert("be_gps");
	return;
    }
    if(sp.phase==2) return;

    // phase 1: set up feature recorders and search sbuf for features
    sbuf_t sbuf = sp.sbuf;
    feature_recorder *exif_recorder = sp.fs.get_name("exif");
    feature_recorder *gps_recorder = sp.fs.get_name("gps");

    // search through sbuf for potential exif content
    size_t start = 0;
    entry_list_t tags;
    while (true) {
cout << "scan_be_exif.b\n";
        // loop to find exif content
        ssize_t exif_start = exif_reader::find_exif_data(sbuf+start);
cout << "scan_be_exif.c" << exif_start << "\n";
        if (exif_start == -1) {
            break;
        }

        // we are at exif content
        sbuf_t exif_sbuf = sbuf + exif_start;

        // process this exif
        // get md5 for this exif
        string md5_hex = md5hex_4k(exif_sbuf);

cout << "scan_be_exif.d\n";
        // get tags for this exif
        exif_reader::read_exif_data(exif_sbuf, tags);

cout << "scan_be_exif.e\n";
        // record to exif feature recorder
        string formatted_exif_data = format_exif_data(tags);
        exif_recorder->write(sp.sbuf.pos0+start, md5_hex, formatted_exif_data);

cout << "scan_be_exif.f\n";
        // record to gps feature recorder
        string formatted_gps_data = format_gps_data(tags);
        gps_recorder->write(sp.sbuf.pos0+start, md5_hex, formatted_gps_data);

cout << "scan_be_exif.g\n";
        // advance start
        start = exif_start + 1;
cout << "scan_be_exif.h\n";
    }
}


inline string format_exif_data(const entry_list_t &tags)
{
    // compose xml from all tags
    stringstream ss;
    ss << "<exif>";
    for (entry_list_t::const_iterator it = tags.begin(); it!=tags.end(); it++) {
        ss << "<" << (*it)->get_full_name() << ">" << xml::xmlescape((*it)->value) << "</" << (*it)->get_full_name() << ">";
    }
    ss << "</exif>";

    return ss.str();
}

inline string format_gps_data(const entry_list_t &tags)
{
    // desired GPS strings
    string gps_time;
    string gps_lon_ref;
    string gps_lon;
    string gps_lat_ref;
    string gps_lat;
    string gps_ele;
    string gps_speed;
    string gps_course;

    // get the desired GPS strings from the tags
    for (entry_list_t::const_iterator it = tags.begin(); it!=tags.end(); it++) {
        if ((*it)->ifd_type == IFD0_GPS) {
            if ((*it)->name == "DateTimeOriginal"){
                gps_time = (*it)->value;
            } else if ((*it)->name == "GPSLongitudeRef"){
                gps_lon_ref = (*it)->value;
            } else if ((*it)->name == "GPSLongitude"){
                gps_lon = (*it)->value;
            } else if ((*it)->name == "GPSLatitudeRef"){
                gps_lat_ref = (*it)->value;
            } else if ((*it)->name == "GPSLatitude"){
                gps_lat = (*it)->value;
            } else if ((*it)->name == "GPSAltitude"){
                gps_ele = (*it)->value;
            } else if ((*it)->name == "GPSSpeed"){
                gps_speed = (*it)->value;
            } else if ((*it)->name == "GPSTrack"){
                gps_course = (*it)->value;
            }
        }
    }

    // compose the formatted gps return value from the desired GPS strings
    return gps_time+","+gps_lat_ref+gps_lat+","
           +gps_lon_ref+gps_lon+","
           +gps_ele+","+gps_speed+","+gps_course;
}

#endif

