/**
 * scan_exif: - requires exiv2 library to be installed.
 * Revision history:
 * 2011-jun-29 slg - Increase exif_gulp_size to 1MB (is this needed at all anymore?)
 *                 - Added support for getting GPS information from photo and adding it to gps.txt
 */

#include "bulk_extractor.h"
#include "xml.h"

#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <iomanip>
#include <cassert>
#include <algorithm>

#include "md5.h"

#ifdef HAVE_EXIV2
/* exiv2 has errors */
#ifdef GNUC_HAS_DIAGNOSTIC_PRAGMA
#pragma GCC diagnostic ignored "-Wshadow"
#pragma GCC diagnostic ignored "-Weffc++"
#endif

#include <exiv2/image.hpp>
#include <exiv2/exif.hpp>
#include <exiv2/error.hpp>

const size_t min_exif_size = 4096;
const size_t exif_gulp_size = 1024*1024;	// how many bytes of EXIF to read

/****************************************************************/
/* C++ string splitting code from http://stackoverflow.com/questions/236129/how-to-split-a-string-in-c */
static std::vector<std::string> &split(const std::string &s, char delim, std::vector<std::string> &elems) {
    std::stringstream ss(s);
    std::string item;
    while(std::getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}


static std::vector<std::string> split(const std::string &s, char delim) {
    std::vector<std::string> elems;
    return split(s, delim, elems);
}




using namespace std;

/** 
 * jpeg_start() returns true if the buffer starts with a jpeg.
 */
inline bool jpeg_start(const sbuf_t &sbuf){
    return (sbuf[0]==(unsigned char)'\xff'
	    && sbuf[1]==(unsigned char)'\xd8'
	    && sbuf[2]==(unsigned char)'\xff');
}

#if 0
static size_t max(size_t a,size_t b)
{
    return a>b ? a : b;
}
#endif

static size_t min(size_t a,size_t b)
{
    return a<b ? a : b;
}

/** Return the hex md5 of the first 4k of a buffer or the entire buffer, whichever is smaller. */
static string md5hex_4k(const sbuf_t &sbuf)
{
    return md5_generator::hash_buf(sbuf.buf,min(sbuf.pagesize,4096)).hexdigest();
}

/**
 * Used for helping to convert libexiv2's time format to ISO8601
 */
static char slash_to_dash(char ch)
{
    if(ch=='/') return '-';
    return ch;
}


/**
 * Used for helping to convert libexiv2's GPS format to decimal lat/long
 */

static double stod(string s)
{
    double d=0;
    sscanf(s.c_str(),"%lf",&d);
    return d;
}

static string dtos(double d)
{
    char buf[64];
    snprintf(buf,sizeof(buf),"%f",d);
    return string(buf);
}

static double rational(string s)
{
    std::vector<std::string> parts = split(s,'/');
    if(parts.size()!=2) return stod(s);	// no slash, so return without
    double top = stod(parts[0]);
    double bot = stod(parts[1]);
    return bot>0 ? top / bot : top;
}

static string fix_gps(string s)
{
    std::vector<std::string> parts = split(s,' ');
    if(parts.size()!=3) return s;	// return the original
    double res = rational(parts[0]) + rational(parts[1])/60.0 + rational(parts[2])/3600.0;
    s = dtos(res);
    return s;
}

static string fix_gps_ref(string s)
{
    if(s=="W" || s=="S") return "-";
    return "";
}

int exif_show_all=1;
extern "C"
void scan_exif(const class scanner_params &sp,const recursion_control_block &rcb)
{
    assert(sp.version==1);
    if(sp.phase==0){
	sp.info->name  = "exif";
	sp.info->feature_names.insert("exif");
	sp.info->feature_names.insert("gps");
	return;
    }
    if(sp.phase==2) return;

    const sbuf_t &sbuf = sp.sbuf;
    feature_recorder *exif_recorder = sp.fs.get_name("exif");
    feature_recorder *gps_recorder = sp.fs.get_name("gps");

#ifdef HAVE_EXIV2__LOGMSG__SETLEVEL
    /* New form to suppress error messages on exiv2 */
    Exiv2::LogMsg::setLevel(Exiv2::LogMsg::mute);
#endif    


    size_t pos_max = 1;		// by default, just scan 1 byte
    if(sbuf.bufsize > min_exif_size){
	pos_max = sbuf.bufsize - min_exif_size; //  we can scan more!
    }
    
    /* Loop through all possible locations in the buffer */
    for(size_t pos=0;pos<pos_max;pos++){
	size_t count = exif_gulp_size;
	count = min(count,sbuf.bufsize-pos);
	//size_t count = sbuf.bufsize-pos; // use all to end

	/* Explore the beginning of each 512-byte block as well as the starting location
	 * of any JPEG on any boundary. This will cause processing of any multimedia file
	 * that Exiv2 recognizes (for which I do not know all the headers.
	 */
	if(pos%512==0 || jpeg_start(sbuf+pos)){
	    try {
		Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open(sbuf.buf+pos,count);
		if(image->good()){
		    image->readMetadata();
		    
		    Exiv2::ExifData &exifData = image->exifData();
		    if (exifData.empty()) continue;
		    
		    /*
		     * Create the MD5 of the first 4K to use as a unique identifier.
		     */
		    string md5_hex = md5hex_4k(sbuf+pos); 

		    char xmlbuf[1024];
		    snprintf(xmlbuf,sizeof(xmlbuf),
			     "<exif><width>%d</width><height>%d</height>",
			     image->pixelWidth(),image->pixelHeight());

		    /* Scan through the fields reported by exiv2.
		     * Create XML for each and extract GPS information if present.
		     */
		    string gps_time;
		    string gps_lat_ref;
		    string gps_lat;
		    string gps_lon_ref;
		    string gps_lon;
		    string gps_ele;
		    string gps_speed;
		    string gps_course;

		    string xml = xmlbuf;
		    for (Exiv2::ExifData::const_iterator i = exifData.begin();
			 i != exifData.end(); ++i) {
//zz testing			if(i->count()>1000) continue;	// ignore long ones
			string key = string(i->key());
			xml.append("<"+key+">"+xml::xmlescape(i->value().toString())+"</"+key+">");
			if(key=="Exif.Photo.DateTimeOriginal"){
			    gps_time = i->value().toString();

			    /* change "2011/06/25 12:20:11" to "2011-06-25 12:20:11" */
			    std::transform(gps_time.begin(),gps_time.end(),gps_time.begin(),slash_to_dash);

			    /* Change "2011-06-25 12:20:11" to "2011-06-25T12:20:11" */
			    size_t first_space = gps_time.find(' ');
			    if(first_space!=string::npos) gps_time.replace(first_space,1,"T");
			}
			if(key=="Exif.GPSInfo.GPSLongitudeRef"){
			    gps_lon_ref = fix_gps_ref(i->value().toString());
			}
			if(key=="Exif.GPSInfo.GPSLongitude"){
			    gps_lon = fix_gps(i->value().toString());
			}
			if(key=="Exif.GPSInfo.GPSLatitudeRef"){
			    gps_lat_ref = fix_gps_ref(i->value().toString());
			}
			if(key=="Exif.GPSInfo.GPSLatitude"){
			    gps_lat = fix_gps(i->value().toString());
			}
			if(key=="Exif.GPSInfo.GPSAltitude"){
			    gps_ele   = dtos(rational(i->value().toString()));
			}
			if(key=="Exif.GPSInfo.GPSSpeed"){
			    gps_speed = dtos(rational(i->value().toString()));
			}
			if(key=="Exif.GPSInfo.GPSTrack"){
			    gps_course = i->value().toString();
			}
		    }
		    exif_recorder->write(sp.sbuf.pos0+pos,md5_hex,xml);
		    gps_recorder->write(sp.sbuf.pos0+pos,md5_hex,
					gps_time+","+gps_lat_ref+gps_lat+","
					+gps_lon_ref+gps_lon+","
					+gps_ele+","+gps_speed+","+gps_course);
		}
	    }
	    //catch (Exiv2::BasicError<char> &e) { }
	    catch (Exiv2::AnyError &e) { }
	    catch (std::exception &e) { }
	}
    }
}

#endif
