%{
/* scan_gps.flex:
 * A flex-based scanner to pick up GPS coordinates.
 * Currently supports Garmin Trackpt format.
 * Note that we could do this with an XML parser, but that wouldn't pick up fragments
 * within memory
 *
 *  http://flex.sourceforge.net/manual/Cxx.html
 */

/*
 * Here are samples that we crib from:
 *
 * A garmin Trackpoint from a Nuvi 240
 *

 <trkpt lat="38.866110" lon="-77.136286"><ele>90.98</ele><time>2010-10-16T20:46:52Z</time><extensions><gpxtpx:TrackPointExtension><gpxtpx:speed>21.96</gpxtpx:speed><gpxtpx:course>87.53</gpxtpx:course></gpxtpx:TrackPointExtension></extensions></trkpt>

 */

#include "bulk_extractor.h"
#include "utils.h"

#include <stdlib.h>
#include <string.h>

/* Needed for flex */
#undef 	ECHO
#define	ECHO {}

#define YY_SKIP_YYWRAP
#define YY_DECL int MyGPSFlexLexer::yylex()
#define YY_INTERACTIVE 1  /* experimental */

class MyGPSFlexLexer : public gpsFlexLexer {
      /* Standards for all flex lexers */
      public:
      MyGPSFlexLexer(const sbuf_t &sbuf_):sbuf(sbuf_),pos(0),point(0),eof(false){
      }
      const sbuf_t &sbuf;
      size_t pos;
      size_t point;
      bool eof;
      virtual int LexerInput(char *buf,int max_size_){  /* signature can't be changed */
      	      if(eof) return 0;
      	      u_int max_size = max_size_ > 0 ? max_size_ : 0;
      	      int count = 0;
      	      if(max_size + point > sbuf.bufsize){
	         max_size = sbuf.bufsize - point;
              }
	      while(max_size>0){
	         *buf = (char)sbuf.buf[point];
		 buf++;
		 point++;
		 max_size--;
		 count++;
  	      }
	      return count;
      }
      virtual int yylex();

      /* specific for GPS module */

      static string get_quoted_attrib(string text,string attrib);
      static string get_cdata(string text);
      void clear();
      class feature_recorder *gps_recorder;
      string lat;
      string lon;
      string ele;
      string time;
      string speed;
      string course;
};

int gpsFlexLexer::yylex(){return 0;}

/**
 * Return NNN in <tag attrib="NNN">
 */

string MyGPSFlexLexer::get_quoted_attrib(string text,string attrib)
{
	size_t pos = text.find(attrib);
	if(pos==string::npos) return "";  /* no attrib */
	ssize_t quote1 = text.find('\"',pos);
	if(quote1<0) return "";           /* no opening quote */
	ssize_t quote2 = text.find('\"',quote1+1);
	if(quote2<0) return "";           /* no closing quote */
	return text.substr(quote1+1,quote2-(quote1+1));
}

/**
 * Return NNN in <tag>NNN</tag>
 */

string MyGPSFlexLexer::get_cdata(string text)
{
	ssize_t gt = text.find('>');
	if(gt<0) return "";           /* no > */
	ssize_t lt = text.find('<',gt+1);
	if(lt<0) return "";           /* no < */
	return text.substr(gt+1,lt-(gt+1));
}

/**
 * dump the current and go to the next
 */
void MyGPSFlexLexer::clear()
{
	if(time.size() || lat.size() || lon.size() || ele.size() || speed.size() || course.size()){
		string what = time+","+lat+","+lon+","+ele+","+speed+","+course;
		gps_recorder->write(sbuf.pos0+pos,what,"");
	}
	time = "";
	lat = "";
	lon = "";
	ele = "";
	speed = "";
	course = "";
}

%}

%option c++
%option noyywrap
%option 8bit
%option batch
%option case-insensitive
%option pointer
%option noyymore
%option prefix="gps"

LATLON	(-?[0-9]{1,3}[.][0-9]{6,8})
ELEV	(-?[0-9]{1,6}[.][0-9]{0,3})

%%

[<]trkpt\ lat=\"{LATLON}\"\ lon=\"{LATLON}\"  {
	clear();
	lat = get_quoted_attrib(yytext,"lat");
	lon = get_quoted_attrib(yytext,"lon");
        pos += yyleng; 
}


[<]ele[>]{ELEV}[<][/]ele[>] {
        ele = get_cdata(yytext);
        pos += yyleng; 
}

[<]time[>][0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9][ T][0-9][0-9]:[0-9][0-9]:[0-9][0-9](Z|([-+][0-9.]))[<][/]time[>] {
        time = get_cdata(yytext);
        pos += yyleng; 
}

[<]gpxtpx:speed[>]{ELEV}[<][/]gpxtpx:speed[>] {
        speed = get_cdata(yytext);
        pos += yyleng;
}

[<]gpxtpx:course[>]{ELEV}[<][/]gpxtpx:course[>] {
        course = get_cdata(yytext);
        pos += yyleng;
}

.|\n { 
     /**
      * The no-match rule.
      * If we are beyond the end of the margin, call it quits.
      */
     /* putchar(yytext[0]); */ /* Uncomment for debugging */

     /* If we have an invalid character, then we are out of this XML block. Clear */
     if(yytext[0] & 0x80){
         clear();
     }

     pos++; 
     if(pos>=sbuf.pagesize) eof=true;
}
%%

extern "C"
void scan_gps(const class scanner_params &sp,const recursion_control_block &rcb)
{
    assert(sp.version==1);      
    if(sp.phase==0){
	sp.info->name  = "gps";
        sp.info->feature_names.insert("gps");
	return;
    }
    if(sp.phase==2) return;

    /* Prescan */
    if(sp.sbuf.find("trkpt",0)==-1 || sp.sbuf.find("lat=",0)==-1 || sp.sbuf.find("lon=",0)==-1) return;


    MyGPSFlexLexer  lexer(sp.sbuf);
    lexer.gps_recorder = sp.fs.get_name("gps");

    while(lexer.yylex() != 0) {
    }
}

