/**
 *
 * ABOUT:
 *	 This is a non-plugin version of identifying and analyzing
 *	prefetch files on Windows for bulk_extractor.
 *
 * AUTHOR:
 *  Luis E. Garcia II
 *  Contact: 
 *
 *
 * NOTE: 
 * See  http://forensicswiki.org/wiki/Prefetch for Prefetch related 
 * information.
 * Anaylzing a raw image of roughly 21GB will take approximately 9 minutes
 *
 * ISSUES:
 * When encountering corrupt data (e.g., sections of file over written), this program will
 * attempt to reduce as much garbage-related output as possible. There may be instances 
 * where some of the captured data will be inaccurate (corrupted data).
 *
 */

#include "bulk_extractor.h"

#include <iostream>
#include <fstream>
#include <string>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <sstream>
#include <vector>

#include "utf8.h"

#ifdef NOT_NEEDED
uint64_t le64toh(uint64_t v) 
{
    uint8_t *b = (uint8_t *)&v;

    return b[0]
	| ((uint64_t)(b[1])<<8) 
	| ((uint64_t)(b[2])<<16) 
	| ((uint64_t)(b[3])<<24) 
	| ((uint64_t)(b[4])<<32) 
	| ((uint64_t)(b[5])<<40) 
	| ((uint64_t)(b[6])<<48) 
	| ((uint64_t)(b[7])<<56);
}
#endif

using namespace std;

class PrefetchDecoder {
private:
    /*** neither copying nor assignment is implemented ***
     *** We do this by making them private constructors that throw exceptions. ***/
    class not_impl: public exception {
	virtual const char *what() const throw() {
	    return "copying feature_recorder objects is not implemented.";
	}
    };
    PrefetchDecoder(const PrefetchDecoder &pd):base_(0),bufsize(0),prefetchFileSize(0),
					       runsOffset(0),lastRanTimeOffset(0),
					       vpOffset(0),vpPivotOffset(0),version(),directories(){
	throw new not_impl();
    }
    const PrefetchDecoder &operator=(const PrefetchDecoder &pd){throw new not_impl();}

    static const uint8_t PREFETCH_HEADER_WINDOWS7[8];
    static const uint8_t PREFETCH_HEADER_XP[8];
    static const uint8_t EXECUTABLE_NAME_OFFSET = 0x10;
    static const size_t  MIN_PREFETCH_SIZE = 1024;
    static const uint64_t SECONDS_BETWEEN_WIN32_EPOCH_AND_UNIX_EPOCH = 11644473600LL;
    static const uint64_t ONE_HUNDRED_NANO_SEC_TO_SECONDS = 10000000;
    static const uint32_t DISTANCE_OFFSET_TO_SEC_LIST = 0x18;  
    static const uint32_t DISTANCE_OFFSET_TO_SERIALNUM = 0x10;
    static const uint32_t DISTANCE_TO_CREATION_TIME = 0x8;

    /*
     * 11644473600 is the number of seconds between the Win32 epoch
     * and the Unix epoch.
     *
     * http://arstechnica.com/civis/viewtopic.php?f=20&t=111992
     */

public:
    /**
     * Service routines 
     */
  
    /* 
     *Byte-order independent cast/conversion;
     * Requires that we can read 4 bytes at buf.
     */
    static string utf16to8(wstring s){
	string utf8_line;
    
	try {
	    utf8::utf16to8(s.begin(),s.end(),back_inserter(utf8_line));
	} catch(utf8::invalid_utf16){
	    /* Exception thrown: bad UTF16 encoding */
	    utf8_line = "";
	}
	return utf8_line;
    }

    static string microsoftDateToISODate(const uint64_t &time) {
	/**
	 * See comment above for more information on
	 * SECONDS_BETWEEN_WIN32_EPOCH_AND_UNIX_EPOCH 
	 *
	 * Convert UNIX time_t to ISO8601 format
	 */
	time_t tmp = (time / ONE_HUNDRED_NANO_SEC_TO_SECONDS)
	    - SECONDS_BETWEEN_WIN32_EPOCH_AND_UNIX_EPOCH;
    
	struct tm time_tm;
	gmtime_r(&tmp, &time_tm);
	char buf[256];
	strftime(buf, sizeof(buf), "%Y-%m-%dT%H:%M:%S", &time_tm);
	return string(buf);
    }

    /** Prefetch constants */

    static const uint32_t headerSizeOffset = 0x54;
    static const uint32_t dNameLengthOffset = 0x4;
    static const uint32_t vpPos = 0x6c;
    static const uint32_t eofOffset = 0xc; // where the end of file is
    static const uint32_t filesOffset = 0x64;
    static const uint32_t filesLengthOffset = 0x68;

    /**
     * Prefetch variables
     *
     * Not too sure if time is saved as signed or unsigned.
     * Being tip-picky
     */

    const uint8_t	*base_;	/* start of prefetch file in mapped file; never access directly! */
    size_t		bufsize;	/* counting from base; don't read past base+bufsize */
    uint32_t		prefetchFileSize;  // stored, but we make it smaller
    uint32_t		runsOffset;		/* Depends on which Windows Version */
    uint32_t		lastRanTimeOffset;/* Depends on which Windows Version */
    uint32_t		vpOffset; /* Volume Path (vp) offset. Depends*/
    uint32_t		vpPivotOffset;
    string		version; 		/* Indicates which version of Windows */
    wstring	directories;

    /* Constructor */
    PrefetchDecoder():base_(0),bufsize(0),prefetchFileSize(0),
		      runsOffset(0),lastRanTimeOffset(0),
		      vpOffset(0),vpPivotOffset(0),version(),directories(){ }
    virtual ~PrefetchDecoder(){ }
	
    /* Return a byte safely from base. */
    uint8_t base(size_t offset) const {
	return offset<bufsize ? base_[offset] : 0;
    }

    /**
     * Return the number of nulls in a region of the buffer
     */

    uint32_t null2_count(wstring str){
	uint32_t count = 0;
	for(wstring::const_iterator it = str.begin(); it!=str.end();it++){
	    if((*it)==0) count++;
	}
	return count;
    }

    /**
     * Return the number of non-ASCII characters in region of the buffer
     */
    uint32_t nonascii_count(wstring str){
	uint32_t count = 0;
	for(wstring::const_iterator it = str.begin(); it!=str.end();it++){
	    if ((*it)>0x7f) count++;
	}
	return count;
    }

    /** 
     * Return the unsigned 32-bit value at a given offset from 
     * the beginning of the buffer, with error checking.
     */
    uint32_t get32i(size_t offset) const {
	if(offset+3 > prefetchFileSize) return 0; // cannot read
	return (base(offset+3)<<24)|(base(offset+2)<<16)|(base(offset+1)<<8)|(base(offset));
    }

    uint64_t get64i(size_t offset) const{
	if(offset+8 > prefetchFileSize) return 0;
	return base(offset+0)
	    | ((uint64_t)(base(offset+1))<<8) 
	    | ((uint64_t)(base(offset+2))<<16) 
	    | ((uint64_t)(base(offset+3))<<24) 
	    | ((uint64_t)(base(offset+4))<<32) 
	    | ((uint64_t)(base(offset+5))<<40) 
	    | ((uint64_t)(base(offset+6))<<48) 
	    | ((uint64_t)(base(offset+7))<<56);

    }

    /** 
     * Return a 16-bit character from an offset in the buffer, with error checking 
     */
    uint16_t get16i(size_t offset) const {
	if(offset+1 > prefetchFileSize) return 0; // asked to read beyond buffer
	return (base(offset+1)<<8) | (base(offset));
    }

    /** 
     * Return a UTF16 string from a given offset in the buffer, with error checking 
     */
    wstring getUTF16(size_t offset,size_t maxChars) const{
	wstring ret;
	for(;maxChars>0;maxChars--){
	    uint16_t ch = get16i(offset);
	    /*test if consistent charset*/
	    if(ch==0) break;
	    if(ch>0) ret.push_back(ch);
	    offset+=2;
	}
	return ret;
    }

    /**
     * Prefetch related operations
     */


    /** 
     * identifyBuf
     * @param buf - buffer to examine to see if it contains the start of a prefetch file.
     * @param bufsize - the size of the buffer
     * @return true - this is a prefetch
     * side effect - analyzes the prefetch file if it one is found
     */
    bool     identifyBuf(const uint8_t *buf,size_t bufsize);

    /**
     * setupPrefetchLayout
     * @param v - Identifies Windows Operating systems
     * @param lRanTimeOffset - Offset to last ran time 
     * @param rOffset - Offset to the number of time an application ran
     * @param fOffset - Offset to file listing
     * @param fLength - Offset to file listing size
     * sets up all offset values to decode prefetch correctly
     */

    void setupPrefetchLayout(string v, uint32_t lRanTimeOffset, uint32_t rOffset){
	version = v;
	runsOffset = rOffset;
	lastRanTimeOffset = lRanTimeOffset;
	vpOffset = get32i(vpPos);
	vpPivotOffset = get32i(vpOffset);
	prefetchFileSize = get32i(eofOffset); // we experimentally determined this is the file length
    }

    /**
     * Retrieval functions
     */

    string lastRunTime() const{
	/* Get the last run time in a byte-order independent manner */
	uint64_t mstime=
	    (((uint64_t)get32i(lastRanTimeOffset+4)) <<32 ) | get32i(lastRanTimeOffset);

	return microsoftDateToISODate(mstime);
    }

    /**
     * Return the name of the exectuable that created the prefetch file.
     * as a UTF16 string.  We experimentally determined that the executable
     * name is termianted with a U+0000 character.
     */
    wstring getExecNameUTF16() const{
	wstring etmp;
	for(size_t look = EXECUTABLE_NAME_OFFSET;
	    (base(look)!=0 || base(look+1)!=0);
	    look+=2){
	    etmp.push_back(base(look) | (base(look+1)<<8));	/* append the character */
	}
	return etmp;
    }
   
    wstring getdNameUTF16() const{
	uint32_t dNameLength =  get32i(vpOffset+dNameLengthOffset);

	/*vpOffset and vpPivotOffset was bound checked when initialized*/
	size_t dNameStart = vpOffset+vpPivotOffset;
	wstring etmp;

	for(size_t pos = 0; 
	    pos < dNameLength*2 && (base(dNameStart+pos) != 0 || (base(dNameStart+pos+1) != 0)); 
	    pos+=2){

	    etmp.push_back(base(dNameStart+pos) | (base(dNameStart+pos+1)<<8) );
	}

	return etmp;
    }

    string getExecNameUTF8() const {
	return utf16to8(getExecNameUTF16());
    }

    string getCreationTime() const {
	uint32_t creationTimeOffset = vpOffset + DISTANCE_TO_CREATION_TIME;	
	uint64_t creationTime = get64i(creationTimeOffset);
	return microsoftDateToISODate(creationTime);
    }

    /*Volume serial number*/
    uint32_t getSerialName() const{
	uint32_t serialOffset = vpOffset+DISTANCE_OFFSET_TO_SERIALNUM;
	return get32i(serialOffset);
    }

    /**
     * Get all the embedded UTF16 strings from Prefetch file
     */
    vector <wstring> pathnameListFromOffset(size_t offset,uint32_t filesLength,uint32_t directorySize)  {
	vector<wstring> ret;		 /* the return list */

	for(size_t pos = offset;pos < prefetchFileSize; ){
	    /* Get the next string */
	    wstring str;
	    if(filesLength) str = getUTF16(pos,filesLength);
	    if(directorySize) str = getUTF16(pos,directorySize);
	    if(str.size()==0) break;	/* no more strings */

	    /* See if the string crosses a 512-byte boundary */
	    size_t bytesIntoBlock = pos % 512;
	    if(bytesIntoBlock + str.size()*2 > 512){
		/**
		 * The string crosses a boundary.
		 */
		uint32_t count_start = nonascii_count(str.substr(0,8)); // number of non-ascii in first 8 bytes
		if(count_start < 4){

		    wstring last8;
		    try{
			/* If the string's start is mostly ASCII */
			last8 = str.substr(str.size()-8,8);
		    }catch(exception& e){
			/**
			 * When the substr accesses an out-of-bound region
			 * an exception is thrown. Once caught
			 * lower the size of substring to capture.
			 */
			last8 = str.substr(str.size()-4,4);
	     
		    }
		    uint32_t count_end = nonascii_count(last8);
		    uint32_t count_nul =  null2_count(last8);

		    if((count_end >= 4) || (count_nul>1)){
			/**
			 * Four (4) or more characters before the block boundary are ASCII,
			 * and fewer than half after the block boundary are ASCII.
			 * The prefetch file was probably truncated.
			 * So make the buffer size equal to the current pos
			 * and return.
			 */
			prefetchFileSize = pos; /* don't look past this transition point */
			break;
		    }
		}
	    }

	    ret.push_back(str);		/* add to the list */

	    /* Number of characters to skip forward is str.size()*2+2 */
	    if(filesLength){
		pos         += str.size()*2+2;
		filesLength -= str.size()*2+2;
	    }
	    if(directorySize){
		pos += str.size()*2+4;
		directorySize -= str.size()*2+4;
	    }
	}
	return ret;
    }

    /**
     * getFilesUTF16() builds a vector of the file names (DLLs mostly)
     * from the prefect file.
     *
     * Also examines the buffer as we iterate through to see if we are at
     * a possible truncation point. This is indicated by a sudden
     * entropy change on a block boundary.
     */

    /* return the first list of files */
    vector<wstring> getFilesUTF16()  {

	/**
	 * Get all associated start-up file paths
	 * The paths are simply packed as a block of strings, each terminated
	 * by a U+0000.
	 **/

	/* Find offset in buffer*/
	uint32_t filesPos = get32i(filesOffset);

	/* Find Length of file paths including NULL bytes */
	uint32_t filesLength = get32i(filesLengthOffset); 

	return pathnameListFromOffset(filesPos,filesLength,0);
    }

    vector<wstring> getDirectoriesUTF16() {
	/**
	 * Could not find any writings stating the size of
	 * the second listing. The offset to the second file
	 * listing was found at 8 (0x08) + location of serial number.
	 * To verify this finding, a set of different Windows
	 * prefetch files were examined. 
	 *
	 * Note: The size of the second file listing is still
	 * unknown. It can be assumed that the EOF will end the second listing.
	 * All prefetch file format end with the second file listing.
	 */

	vector<wstring> ret;

	uint32_t dNameLength =  get32i(vpOffset+dNameLengthOffset);/* This assumes 4-byte alignment */
	uint32_t quot = dNameLength % 4;
	uint32_t alignDName = 0; 
	if(quot != 0){
	    alignDName += (4 - quot);	
	}
	alignDName += dNameLength;

	/**
	 * This will get the distance value between end of volume path string 
	 * to second set of strings
	 */
	uint32_t dirListPos = vpOffset+DISTANCE_OFFSET_TO_SEC_LIST;
	uint32_t dirListDistance = get32i(dirListPos);	/*safety-check*/

	uint32_t endOfVolPathName = vpOffset+vpPivotOffset+alignDName*2;
	uint32_t startOfDirList = dirListDistance+endOfVolPathName+0x2;
	uint32_t eof = get32i(eofOffset);
	uint32_t directorySize = eof-startOfDirList;

	return pathnameListFromOffset(startOfDirList,0,directorySize);
    }

    /* bytes order dependent */
    uint32_t getHeaderSize() const{return get32i(headerSizeOffset);}
    uint32_t getRuns() const{return get32i(runsOffset);}
    uint32_t getPrefetchFileSize() const{ return prefetchFileSize;}

    /*
     * XML generation. May be pretty-printed, may not be.
     */
    string toXML(bool pp) {		
	/**
	 * XML UTF-8 output
	 * Format: http://www.forensicswiki.org/wiki/Prefetch_XML
	 **/

	/* First search for files and directories; this will detect truncation and resize buffer */
	vector<wstring> files1 = getFilesUTF16();
	vector<wstring> dirs1  = getDirectoriesUTF16();

	stringstream ss;

	/* These string constants are used for pretty printing */
	string nl = pp ? "\n" : "";
	string sp2 = pp ? "  " : "";
	string sp4 = pp ? "  " : "";

	ss << "<prefetch>" << nl 
	   << sp2 << "<header>" << nl 
	   << sp4 << "<os>" << version << "</os>" << nl 
	   << sp4 << "<header_size>" << getHeaderSize() << "</header_size>" << nl 
	   << sp4 << "<filename>" << getExecNameUTF8() << "</filename>" << nl 
	   << sp4 << "<runs>" << getRuns() << "</runs>" << nl 
	   << sp4 << "<atime>" << lastRunTime() << "</atime>" << nl 
	   << sp2 << "</header>" << nl 
	   << sp2 << "<volume>" << nl 
	   << sp4 << "<path>" << utf16to8(getdNameUTF16()) << "</path>" << nl 
	   << sp4 << "<serial_number>" << hex << getSerialName() << dec << "</serial_number>" << nl 
	   << sp2 << "</volume>" << nl 
	   << sp2 << "<creation>" << getCreationTime() << "</creation>" << nl 
	   << sp2 << "<filenames>" << nl ;

	for(vector<wstring>::const_iterator it = files1.begin();
	    it != files1.end(); it++){
	    ss << sp4 << "<file>" << utf16to8(*it)  << "</file>" << nl;
	}

	ss << sp2 << "</filenames>" << nl
	   << sp2 << "<dirnames>" << nl;

	for(vector<wstring>::const_iterator it = dirs1.begin();
	    it != dirs1.end(); it++){
	    ss << sp4 << "<dir>" << utf16to8(*it) << "</dir>" << nl;
	}

	ss << sp2 << "</dirnames>" << nl
	   << "</prefetch>" << nl;

	return ss.str();
    }
    void record(const pos0_t &pos0,feature_recorder *pf_recorder)  {
	pf_recorder->write(pos0,this->toXML(false),"");
    }
};

const uint8_t PrefetchDecoder::PREFETCH_HEADER_WINDOWS7[8] = {0x17, 0x00, 0x00, 0x00, 0x53, 0x43, 0x43, 0x41};  
const uint8_t PrefetchDecoder::PREFETCH_HEADER_XP[8]       = {0x11, 0x00, 0x00, 0x00, 0x53, 0x43, 0x43, 0x41};


/**
 * See if buf is a prefetch and return true if it is, false otherwise.
 * If it is, set the required instance variables to determine which type.
 */
bool PrefetchDecoder::identifyBuf(const uint8_t *buf,size_t bufsize_)
{
    /**
     * Read data and perform tests to see if file is a prefetch file
     *
     * The first byte of header is not unique for all Windows installation. 
     * WinXP has \x11 while Win7 has \x17
     *
     * This may need improvement to handle bulk data as for now it takes a file
     * from the command-line 
     *
     * Corrected Issue: when comparing the comparsion operation is being 
     * terminated early because of the \x00 (NULL) characters in the string. 
     * This causes an issue when reading stream of bytes
     */

    if(bufsize_ < MIN_PREFETCH_SIZE){
	return false; /* too small */
    }

    this->base_   = buf;		/* buf is the start of the prefetch file */
    this->bufsize = bufsize; /* amount of bytes we are allowed to analyze; we may shorten it. */

    if(!memcmp(PREFETCH_HEADER_WINDOWS7, buf, sizeof(PREFETCH_HEADER_WINDOWS7))){
	setupPrefetchLayout("Windows Vista or Windows 7", 0x80, 0x98);
	return true;
    }
    if(!memcmp(PREFETCH_HEADER_XP, buf, sizeof(PREFETCH_HEADER_XP))){
	setupPrefetchLayout("Windows XP", 0x78, 0x90);
	return true;
    }
    return false;
}


/* iostream support for PrefetchDecoder */
#ifdef NOT_USED
static std::ostream & operator << (std::ostream &os, class PrefetchDecoder &p)
{
    os << "Prefetch " << p.version << "\n"
       << "HeaderSize: " << p.getHeaderSize() << "\n"
       << "Filesize: " << p.getPrefetchFileSize() << " bytes \n"
       << "Executable Name:" << PrefetchDecoder::utf16to8(p.getExecNameUTF16()) << "\n"
       << "Number of runs: " << p.getRuns() << "\n"
       << "Last Run Time UTC: " << p.lastRunTime() << "\n"
       << "Volume Path: " << PrefetchDecoder::utf16to8(p.getdNameUTF16()) << "\n"
       << "Volume Serial Number: " << hex << p.getSerialName() << dec << "\n"
       << "Creation Time: " << p.getCreationTime() << "\n"
       << "XML:\n" << p.toXML(true) << "\n";

    return os;
}
#endif

extern "C"
void scan_winprefetch(const class scanner_params &sp,const recursion_control_block &rcb)
{
    assert(sp.version==1);
    if(sp.phase==0){
	sp.info->name  = "winprefetch";
	sp.info->feature_names.insert("winprefetch");
	return;
    }
    if(sp.phase==2) return;		// no shutdown
    
    const sbuf_t &sbuf = sp.sbuf;
    feature_recorder_set &fs = sp.fs;
    feature_recorder *pf_recorder = fs.get_name("winprefetch");
    pf_recorder->set_quoting_enabled(false); // because we are sending through XML
    
    /* Look for the signature at the beginning of a prefetch file */
    PrefetchDecoder pfile;
    for(size_t pos=0;pos<sbuf.bufsize;pos++){
	/**
	 * As a prefetch is identified reduce the size of buffer prevent reading too much data
	 */
	if(pfile.identifyBuf(sbuf.buf+pos,sbuf.bufsize-pos)){
	    if(pfile.getExecNameUTF8().size()>0){ 
		pfile.record(sbuf.pos0+pos,pf_recorder);
	    }
	}
    }
}


