#!/usr/bin/env python
# coding=UTF-8
"""
Regression system: 
 - Run bulk_extractor on one of three disk images. 
 - The feature files are sorted (to deal with multi-threading issues).
 - Total number of features are reported and compared with the archives.
"""

b'This module needs Python 2.7 or later.'
from subprocess import Popen,call,PIPE
import os,os.path,glob,sys,zipfile

sys.path.append("python")
sys.path.append(os.getenv("DOMEX_HOME")+"/src/bulk_extractor/trunk/python")

default_infile    = "nps-2009-ubnist1/ubnist1.gen3.raw"
canon_infile      = "nps-2009-canon2/nps-2009-canon2-gen6.raw"
domexusers_infile = "nps-2009-domexusers/nps-2009-domexusers.raw"

import bulk_extractor


def find_file(fn):
    if os.path.exists(fn): return fn
    for ne in ('E01','aff','000','001'):
        tfn = fn.replace(".raw",ne)
        if os.path.exists(tfn): return tfn
    raise RuntimeError("Cannot find file "+fn)


if sys.version_info < (2,7):
    raise "Requires Python 2.7 or above"

def analyze(fn):
    """Print statistics about an output directory"""
    print("Analyze {}".format(fn))

    b = bulk_extractor.BulkReport(fn)
    print("bulk_extractor version: {}".format(b.version()))
    print("Filename:               {}".format(b.imagefile()))
    
    # Determine if any pages were not analyzed
    proc = dict()
    for work_start in b.xmldoc.getElementsByTagName("debug:work_start"):
        threadid = work_start.getAttribute('threadid')
        pos0     = work_start.getAttribute('pos0')
        if pos0 in proc:
            print("*** error: pos0={} was started by threadid {} and threadid {}".format(pos0,proc[pos0],threadid))
        else:
            proc[pos0] = threadid
    for work_end in b.xmldoc.getElementsByTagName("debug:work_end"):
        threadid = work_end.getAttribute('threadid')
        pos0     = work_end.getAttribute('pos0')
        if pos0 not in proc:
            print("*** error: pos0={} was ended by threadid {} but never started!".format(pos0,threadid))
        elif threadid!=proc[pos0]:
            print("*** error: pos0={} was ended by threadid {} but ended by threadid {}".format(pos0,proc[pos0],threadid))
        else:
            del proc[pos0]
    
    for (pos0,threadid) in proc.items():
        print("*** error: pos0={} was started by threadid {} but never ended".format(pos0,threadid))
    

    # Print which scanners were run and how long they took
    scanner_times = []
    scanners = b.xmldoc.getElementsByTagName("scanner_times")[0]
    total = 0
    for path in scanners.getElementsByTagName("path"):
        name    = path.getElementsByTagName("name")[0].firstChild.wholeText
        calls   = int(path.getElementsByTagName("calls")[0].firstChild.wholeText)
        seconds = float(path.getElementsByTagName("seconds")[0].firstChild.wholeText)
        total   += seconds
        scanner_times.append((name,calls,seconds))
    print("Scanner paths by time and calls")
    scanner_times.sort(key=lambda a:a[2],reverse=True)

    print("  {0:>25}  {1:8}  {2:12}  {3:12}  {4:5}".format("name","calls","sec","sec/call","% total"))
    for (name,calls,seconds) in scanner_times:
        print("  {:>25}  {:8.0f}  {:12.4f}  {:12.4f}  {:5.2f}%".format(
                name,calls,seconds,seconds/calls,100.0*seconds/total))
    
    

    hfns = list(b.histograms())
    print("Histogram Files:        {}".format(len(hfns)))
    for fn in sorted(hfns):
        h = b.read_histogram(fn)
        firstline = b.open(fn).readline().strip()
        print("  {:>25} entries: {:>10,}  (top: {})".format(fn,len(h),firstline))

    print("")
    ffns = list(b.feature_files())
    print("Feature Files:        {}".format(len(ffns)))
    for fn in sorted(ffns):
        lines = 0
        for line in b.open(fn):
            lines += 1
        print("  {:>25} lines: {:>10,}".format(fn,lines))
    

def make_zip(dname):
    archive_name = dname+".zip"
    b = bulk_extractor.BulkReport(dname)
    z = zipfile.ZipFile(archive_name,compression=zipfile.ZIP_DEFLATED,mode="w")
    print("Creating ZIP archive {}".format(archive_name))
    for fname in b.all_files:
        print("  adding {} ...".format(fname))
        z.write(os.path.join(dname,fname),arcname=os.path.basename(fname))
    

def make_outdir(outdir_base):
    counter = 1
    while True:
        outdir = outdir_base + ("-%02d" % counter)
        if not os.path.exists(outdir):
            return outdir
        counter += 1

def run(oargs):
    outdir = make_outdir(args.outdir)
    cargs=['-o',outdir]
    if args.jobs: cargs += ['-j'+str(args.jobs)]
    
    cargs += ['-e','all']    # enable all scanners
    #cargs += ['-e','wordlist']    # enable all scanners
    #cargs += ['-e','aes']    # enable all scanners
    if args.extra:     cargs += [args.extra]

    if args.debug: cargs += ['-d'+str(args.debug)]
    cargs += ['-r','tests/alert_list.txt']
    cargs += ['-w','tests/stop_list.txt']
    cargs += ['-w','tests/stop_list_context.txt']
    cargs += ['-f','[a-z\.0-9]*@gsa.gov']
    cargs += ['-F','tests/find_list.txt']
    
    cargs += [args.image]
    print(" ".join(cargs))

    # Now that we have a command, run it under GDB

    if args.gdb:
        with open("/tmp/cmds","w") as f:
            f.write("b malloc_error_break\n")
            f.write("run ")
            f.write(" ".join(cargs))
            f.write("\n")
            
        cmd = ['gdb','-e',args.exe,'-x','/tmp/cmds']
    else:
        print("grunt")
        cmd = [args.exe] + cargs
    print(" ".join(cmd))
    r = call(cmd)
    if r!=0:
        raise RuntimeError("{} crashed with error code {}".format(args.exe,r))
    print("Now sorting files in "+outdir)
    for fn in glob.glob(outdir + "/*.txt"):
        if "histogram" in fn: continue
        fns  = fn+".sorted"
        print("Sorting "+fn)
        os.environ['LC_ALL']='C' # make sure we sort in C order
        call(['sort',fn],stdout=open(fns,"w"))
        call(['mv',fns,fn])

    analyze(outdir)
    # Look for a member of the alert_list in the ALERT list.
    # Look for a member of the stop list in stopped
    return outdir
             
def check(fn,lines):
    found_lines = len(file(fn).read().split("\n"))-1
    if lines!=found_lines:
        print("{:10} expected lines: {} found lines: {}".format(fn,found_lines,lines))

def invalid_line(line):
    if not line: return False    # empty lines are okay
    if line[0]=='#': return False # comments are okay
    if line[0:2]=='n=': return False     # currently all histograms are okay
    if line.count('\t')==2: return False # correct number of tabs
    return True
    

def validate(outdir):
    import glob,os.path
    res = {}
    for fn in glob.glob(outdir+"/*"):
        if fn.endswith('.xml'):
            continue
        if fn.endswith('.dmp'):
            continue
        lines = open(fn).read().split("\n")
        bfn = os.path.basename(fn)
        inv_lines = filter(invalid_line,lines)
        if inv_lines:
            print("{} Invalid lines: {}".format(fn,len(inv_lines)))
            for i in inv_lines:
                print("   ",i)
        #res[bfn] = lines
        #if expected[bfn] != res[bfn]:
        #    print("{:10}: expected {:5} ; got {:5}; off by {:6.3}%".format(
        #            bfn,expected[bfn],res[bfn],(res[bfn]-expected[bfn])*100.0/float(expected[bfn])))

            
            
def diff(dname1,dname2):
    args.max = int(args.max)
    def files_in_dir(dname):
        return [fn.replace(dname+"/","") for fn in glob.glob(dname+"/*")]
    def lines_to_set(fn):
        return set(open(fn).read().split("\n"))
    
    files1 = set(files_in_dir(dname1))
    files2 = set(files_in_dir(dname2))
    if files1.difference(files2):
        print("Files only in {}:\n   {}".format(dname1," ".join(files1.difference(files2))))
    if files2.difference(files1):
        print("Files only in {}:\n   {}".format(dname2," ".join(files2.difference(files1))))

    # Look at the common files
    common = files1.intersection(files2)
    for fn in common:
        def print_diff(dname,prefix,diffset):
            if not diffset:
                return
            print("Only in {}: {} lines".format(dname,len(diffset)))
            count = 0
            for line in sorted(diffset):
                print("{}{}".format(prefix,line))
                count += 1
                if count>args.max:
                    print(" ... +{} more lines".format(len(diffset)-int(args.max)))
                    return

        print("regressdiff {} {}:".format(os.path.join(dname1,fn),os.path.join(dname2,fn)))
        lines1 = lines_to_set(os.path.join(dname1,fn))
        lines2 = lines_to_set(os.path.join(dname2,fn))
        print_diff(dname1,"<",lines1.difference(lines2))
        print_diff(dname2,">",lines2.difference(lines1))
        if lines1 != lines2:
            print("")



if __name__=="__main__":
    import argparse 
    global args
    import sys,time

    parser = argparse.ArgumentParser(description="Perform regression testing on bulk_extractor")
    parser.add_argument("--gdb",help="run under gdb",action="store_true")
    parser.add_argument("--debug",help="specify debug level",type=int)
    parser.add_argument("--corp",help="regress entire corpus",action="store_true")
    parser.add_argument("--outdir",help="specifies output directory",default="regress")
    parser.add_argument("--exe",help="Executable to run",default="src/bulk_extractor")
    parser.add_argument("--image",
                      help=("Specifies image to regress (default %s)" % default_infile),
                      default=default_infile)
    parser.add_argument("--jobs",help="Specifies number of jobs",type=int)
    parser.add_argument("--fast",help="Run with the Canon set",action="store_true")
    parser.add_argument("--full",help="Run with the Domexusers set",action="store_true")
    parser.add_argument("--extra",help="Specify extra arguments")
    parser.add_argument("--gprof",help="Recompile and run with gprof",action="store_true")
    parser.add_argument("--diff",help="diff mode. Compare two outputs",action="store_true")
    parser.add_argument("--max",help="Maximum number of differences to display",default="5")
    parser.add_argument("--memdebug",help="Look for memory errors",action="store_true")
    parser.add_argument("--analyze",help="Specifies a bulk_extractor output file to analyze")
    parser.add_argument("--save",help="Saves analysis for this image in regress.db",action="store_true")
    parser.add_argument("--zip",help="Create a zip archive of the report")
    parser.add_argument("--validate",help="Validate the contents of a report (do not run bulk_extractor)")

    args = parser.parse_args()

    # these are mostly for testing
    if args.validate: validate(args.validate); exit(0)
    if args.analyze: analyze(args.analyze); exit(0)
    if args.zip    : make_zip(args.zip); exit(0)

    if not os.path.exists(args.exe):
        raise RuntimeError("{} does not exist".format(args.exe))

    # Find the bulk_extractor version and add it to the outdir
    version = Popen([args.exe,'-V'],stdout=PIPE).communicate()[0].decode('utf-8').split(' ')[1].strip()
    args.outdir += "-"+version

    drives = os.getenv("DOMEX_CORP") + "/drives/nps/"

    if args.fast:
        args.image  = find_file(drives + canon_infile)
        args.outdir += "-fast"
    elif args.full:
        args.image  = find_file(drives+domexusers_infile)
        args.outdir += "-full"
    else:
        args.image  = find_file(drives+default_infile)
        

    if not args.fast and not args.full:
        args.outdir += "-norm"

    if args.memdebug:
        fn = "/usr/lib/libgmalloc.dylib"
        if os.path.exist(fn):
            print("Debugging with libgmalloc")
            os.putenv("DYLD_INSERT_LIBRARIES",fn)
            os.putenv("MALLOC_PROTECT_BEFORE","1")
            os.putenv("MALLOC_FILL_SPACE","1")
            os.putenv("MALLOC_STRICT_SIZE","1")
            os.putenv("MALLOC_CHECK_HEADER","1")
            os.putenv("MALLOC_PERMIT_INSANE_REQUESTS","1")
            run(args)
            exit(0)

        print("Debugging with standard Malloc")
        os.putenv("MallocLogFile","malloc.log")
        os.putenv("MallocGuardEdges","1")
        os.putenv("MallocStackLogging","1")
        os.putenv("MallocStackLoggingDirectory",".")
        os.putenv("MallocScribble","1")
        os.putenv("MallocCheckHeapStart","10")
        os.putenv("MallocCheckHeapEach","10")
        os.putenv("MallocCheckHeapAbort","1")
        os.putenv("MallocErrorAbort","1")
        os.putenv("MallocCorruptionAbort","1")
        run(args)
        exit(0)

    if args.gprof:
        call(['make','clean'])
        call(['make','CFLAGS=-pg','CXXFLAGS=-pg','LDFLAGS=-pg'])
        outdir = run(args)
        call(['gprof',program,"gmon.out"],stdout=open(outdir+"/GPROF.txt","w"))

    if args.corp:
        outdir_base = "/Volumes/Drobo1/bulk_extractor_regression"
        for (dirpath,dirnames,filenames) in os.walk(os.getenv("DOMEX_CORP")):
            for filename in filenames:
                (root,ext) = os.path.splitext(filename)
                if filename in ['aff','E01','.001']:
                    args.image = os.path.join(dirpath,filename)
                    args.outdir = outdir_base + "/"+os.path.basename(fn)
                    run(args)
        exit(0)

    if args.diff:
        if len(args)!=2:
            raise ValueError("--diff requires two arguments")
        diff(args[0],args[1])
        exit(0)

    run(args)
    validate(args.outdir)

                    
        
